@@ -58,10 +58,6 @@ static char rcsid[] = "$NetBSD: mount_nfs.c,v 1.12.4.1 1996/05/25 22:48:05 fvdl
 #include <sys/stat.h>
 #include <syslog.h>
 
-#include <rpc/rpc.h>
-#include <rpc/pmap_clnt.h>
-#include <rpc/pmap_prot.h>
-
 #ifdef ISO
 #include <netiso/iso.h>
 #endif
@@ -77,6 +73,7 @@ static char rcsid[] = "$NetBSD: mount_nfs.c,v 1.12.4.1 1996/05/25 22:48:05 fvdl
 #include <nfs/nfs.h>
 #undef _KERNEL
 
+#include <netinet/in.h>
 #include <arpa/inet.h>
 
 #include <ctype.h>
@@ -142,7 +139,7 @@ struct nfs_args nfsdefargs = {
 	NFS_ARGSVERSION,
 	NULL,
 	sizeof (struct sockaddr_in),
-	SOCK_DGRAM,
+	SOCK_STREAM,
 	0,
 	NULL,
 	0,
@@ -163,20 +160,12 @@ struct nfs_args nfsdefargs = {
 	0
 };
 
-struct nfhret {
-	u_long		stat;
-	long		vers;
-	long		auth;
-	long		fhsize;
-	u_char		nfh[NFSX_V3FHMAX];
-};
 #define	DEF_RETRY	10000
 #define	BGRND	1
 #define	ISBGRND	2
 int retrycnt;
 int opflags = 0;
 int nfsproto = IPPROTO_UDP;
-int mnttcp_ok = 1;
 u_short port_no = 0;
 int force2 = 0;
 int force3 = 0;
@@ -202,8 +191,6 @@ struct	iso_addr *iso_addr __P((const char *));
 #endif
 void	set_rpc_maxgrouplist __P((int));
 __dead	void usage __P((void));
-int	xdr_dir __P((XDR *, char *));
-int	xdr_fh __P((XDR *, struct nfhret *));
 
 int
 main(argc, argv)
@@ -341,8 +328,10 @@ main(argc, argv)
 			}
 			if (altflags & ALTF_RDIRPLUS)
 				nfsargsp->flags |= NFSMNT_RDIRPLUS;
-			if (altflags & ALTF_MNTUDP)
-				mnttcp_ok = 0;
+			if (altflags & ALTF_MNTUDP) {
+				nfsargsp->sotype = SOCK_DGRAM;
+				nfsproto = IPPROTO_UDP;
+			}
 			if (altflags & ALTF_RESVPORT)
 				nfsargsp->flags |= NFSMNT_RESVPORT;
 #ifdef ISO
@@ -418,7 +407,8 @@ main(argc, argv)
 			nfsargsp->flags |= NFSMNT_RETRANS;
 			break;
 		case 'U':
-			mnttcp_ok = 0;
+			nfsargsp->sotype = SOCK_DGRAM;
+			nfsproto = IPPROTO_UDP;
 			break;
 		default:
 			usage();
@@ -547,7 +537,6 @@ getnfsargs(spec, nfsargsp)
 	char *spec;
 	struct nfs_args *nfsargsp;
 {
-	register CLIENT *clp;
 	struct hostent *hp;
 	static struct sockaddr_in saddr;
 #ifdef ISO
@@ -555,15 +544,10 @@ getnfsargs(spec, nfsargsp)
 	struct iso_addr *isop;
 	int isoflag = 0;
 #endif
-	struct timeval pertry, try;
-	enum clnt_stat clnt_stat;
-	int so = RPC_ANYSOCK, i, nfsvers, mntvers, orgcnt;
 	char *hostp, *delimp;
 #ifdef NFSKERB
 	char *cp;
 #endif
-	u_short tport;
-	static struct nfhret nfhret;
 	static char nam[MNAMELEN + 1];
 
 	strlcpy(nam, spec, sizeof(nam));
@@ -636,112 +620,10 @@ getnfsargs(spec, nfsargsp)
 	}
 #endif /* NFSKERB */
 
-	if (force2) {
-		nfsvers = NFS_VER2;
-		mntvers = RPCMNT_VER1;
-	} else if (force3) {
-		nfsvers = NFS_VER3;
-		mntvers = RPCMNT_VER3;
-	} else {
-		nfsvers = NFS_VER4;
-		mntvers = 0;
-	}
-	orgcnt = retrycnt;
-tryagain:
-	nfhret.stat = EACCES;	/* Mark not yet successful */
-	while (nfsvers < NFS_VER4 && retrycnt > 0) {
-		saddr.sin_family = AF_INET;
-		saddr.sin_port = htons(PMAPPORT);
-		if ((tport = port_no ? port_no : pmap_getport(&saddr,
-		    RPCPROG_NFS, nfsvers, nfsargsp->sotype == SOCK_STREAM ?
-		    IPPROTO_TCP : IPPROTO_UDP)) == 0) {
-			if ((opflags & ISBGRND) == 0)
-				clnt_pcreateerror("NFS Portmap");
-		} else {
-			saddr.sin_port = 0;
-			pertry.tv_sec = 10;
-			pertry.tv_usec = 0;
-			if (mnttcp_ok && nfsargsp->sotype == SOCK_STREAM)
-			    clp = clnttcp_create(&saddr, RPCPROG_MNT, mntvers,
-				&so, 0, 0);
-			else
-			    clp = clntudp_create(&saddr, RPCPROG_MNT, mntvers,
-				pertry, &so);
-			if (clp == NULL) {
-				if ((opflags & ISBGRND) == 0)
-					clnt_pcreateerror("Cannot MNT PRC");
-			} else {
-				clp->cl_auth = authunix_create_default();
-				try.tv_sec = 10;
-				try.tv_usec = 0;
-				if (nfsargsp->flags & NFSMNT_KERB)
-				    nfhret.auth = RPCAUTH_KERB4;
-				else
-				    nfhret.auth = RPCAUTH_UNIX;
-				nfhret.vers = mntvers;
-				clnt_stat = clnt_call(clp, RPCMNT_MOUNT,
-				    xdr_dir, spec, xdr_fh, &nfhret, try);
-				if (clnt_stat != RPC_SUCCESS) {
-					if (clnt_stat == RPC_PROGVERSMISMATCH) {
-						if (nfsvers == NFS_VER3 &&
-						    !force3) {
-							retrycnt = orgcnt;
-							nfsvers = NFS_VER2;
-							mntvers = RPCMNT_VER1;
-							nfsargsp->flags &=
-								~NFSMNT_NFSV3;
-							goto tryagain;
-						} else {
-							fprintf(stderr, "%s",
-							    clnt_sperror(clp,
-								"MNT RPC"));
-						}
-					}
-					if ((opflags & ISBGRND) == 0)
-						warnx("%s", clnt_sperror(clp,
-						    "bad MNT RPC"));
-				} else {
-					auth_destroy(clp->cl_auth);
-					clnt_destroy(clp);
-					retrycnt = 0;
-				}
-			}
-		}
-		if (--retrycnt > 0) {
-			if (opflags & BGRND) {
-				opflags &= ~BGRND;
-				if ((i = fork())) {
-					if (i == -1)
-						err(1, "nqnfs 2");
-					exit(0);
-				}
-				(void) setsid();
-				(void) close(STDIN_FILENO);
-				(void) close(STDOUT_FILENO);
-				(void) close(STDERR_FILENO);
-				(void) chdir("/");
-				opflags |= ISBGRND;
-			}
-			sleep(60);
-		}
-	}
-	if (nfsvers >= NFS_VER4) {
-		saddr.sin_family = AF_INET;
-		saddr.sin_port = htons(NFS_PORT);
-		nfsargsp->fh = spec;
-		nfsargsp->fhsize = strlen(spec) + 1;
-	} else {
-		if (nfhret.stat) {
-			if (opflags & ISBGRND)
-				exit(1);
-			errno = nfhret.stat;
-			warnx("can't access %s: %s", spec, strerror(nfhret.stat));
-			return (0);
-		}
-		saddr.sin_port = htons(tport);
-		nfsargsp->fh = nfhret.nfh;
-		nfsargsp->fhsize = nfhret.fhsize;
-	}
+	saddr.sin_family = AF_INET;
+	saddr.sin_port = htons(NFS_PORT);
+	nfsargsp->fh = spec;
+	nfsargsp->fhsize = strlen(spec) + 1;
 #ifdef ISO
 	if (isoflag) {
 		nfsargsp->addr = (struct sockaddr *) &isoaddr;
@@ -756,59 +638,6 @@ tryagain:
 	return (1);
 }
 
-/*
- * xdr routines for mount rpc's
- */
-int
-xdr_dir(xdrsp, dirp)
-	XDR *xdrsp;
-	char *dirp;
-{
-	return (xdr_string(xdrsp, &dirp, RPCMNT_PATHLEN));
-}
-
-int
-xdr_fh(xdrsp, np)
-	XDR *xdrsp;
-	register struct nfhret *np;
-{
-	register int i;
-	long auth, authcnt, authfnd = 0;
-
-	if (!xdr_u_long(xdrsp, &np->stat))
-		return (0);
-	if (np->stat)
-		return (1);
-	switch (np->vers) {
-	case 1:
-		np->fhsize = NFSX_V2FH;
-		return (xdr_opaque(xdrsp, (caddr_t)np->nfh, NFSX_V2FH));
-	case 3:
-		if (!xdr_long(xdrsp, &np->fhsize))
-			return (0);
-		if (np->fhsize <= 0 || np->fhsize > NFSX_V3FHMAX)
-			return (0);
-		if (!xdr_opaque(xdrsp, (caddr_t)np->nfh, np->fhsize))
-			return (0);
-		if (!xdr_long(xdrsp, &authcnt))
-			return (0);
-		for (i = 0; i < authcnt; i++) {
-			if (!xdr_long(xdrsp, &auth))
-				return (0);
-			if (auth == np->auth)
-				authfnd++;
-		}
-		/*
-		 * Some servers, such as DEC's OSF/1 return a nil authenticator
-		 * list to indicate RPCAUTH_UNIX.
-		 */
-		if (!authfnd && (authcnt > 0 || np->auth != RPCAUTH_UNIX))
-			np->stat = EAUTH;
-		return (1);
-	};
-	return (0);
-}
-
 __dead void
 usage()
 {
