@@ -1,4 +1,5 @@
-/*	$Id: nfs4_bio.c,v 1.5 2002/08/20 21:49:26 rees Exp $	*/
+
+/*	$Id: nfs4_bio.c,v 1.11 2003/06/20 17:05:11 rees Exp $	*/
 /*	$OpenBSD: nfs_bio.c,v 1.35 2002/02/08 08:20:49 csapuntz Exp $	*/
 /*	$NetBSD: nfs_bio.c,v 1.25.4.2 1996/07/08 20:47:04 jtc Exp $	*/
 
@@ -67,16 +68,14 @@
 
 #include <sys/param.h>
 #include <sys/systm.h>
+#include <sys/buf.h>
+#include <sys/kernel.h>
+#include <sys/malloc.h>
+#include <sys/mount.h>
+#include <sys/proc.h>
 #include <sys/resourcevar.h>
 #include <sys/signalvar.h>
-#include <sys/proc.h>
-#include <sys/buf.h>
 #include <sys/vnode.h>
-#include <sys/mount.h>
-#include <sys/kernel.h>
-#include <sys/namei.h>
-
-#include <uvm/uvm_extern.h>
 
 #include <nfs/rpcv2.h>
 #include <nfs/nfsproto.h>
@@ -90,14 +89,11 @@
 #include <nfs4/nfs4node.h>
 #include <nfs4/nfs4_var.h>
 
-#ifdef RPCCLNT
 /* needed for the rpctask definition */
 #include <rpc/rpcclnt.h>
-#endif
 
 extern struct proc *nfs_iodwant[NFS_MAXASYNCDAEMON];
 extern int nfs_numasync;
-struct nfsstats nfsstats;
 
 /*
  * Vnode op for read using bio
@@ -131,7 +127,7 @@ nfs_bioread(vp, uio, ioflag, cred)
 	p = uio->uio_procp;
 	if ((nmp->nm_flag & (NFSMNT_NFSV3 | NFSMNT_GOTFSINFO)) == NFSMNT_NFSV3)
 		(void)nfs_fsinfo(nmp, vp, cred, p);
-	biosize = nmp->nm_rsize;
+	biosize = nmp->nm_rpcclnt.rc_rsize;
 	/*
 	 * For nfs, cache consistency can only be maintained approximately.
 	 * Although RFC1094 does not specify the criteria, the following is
@@ -166,7 +162,6 @@ nfs_bioread(vp, uio, ioflag, cred)
 		}
 	}
 
-#ifndef NFSV4
 	/*
 	 * update the cache read creds for this vnode
 	 */
@@ -174,7 +169,6 @@ nfs_bioread(vp, uio, ioflag, cred)
 		crfree(np->n_rcred);
 	np->n_rcred = cred;
 	crhold(cred);
-#endif
 
 	do {
 	    if ((vp->v_flag & VROOT) && vp->v_type == VLNK) {
@@ -376,7 +370,6 @@ nfs_write(v)
 		return (EFBIG);
 	}
 
-#ifndef NFSV4
 	/*
 	 * update the cache write creds for this node.
 	 */
@@ -384,14 +377,13 @@ nfs_write(v)
 		crfree(np->n_wcred);
 	np->n_wcred = cred;
 	crhold(cred);
-#endif
 
 	/*
 	 * I use nm_rsize, not nm_wsize so that all buffer cache blocks
 	 * will be the same size within a filesystem. nfs_writerpc will
 	 * still use nm_wsize when sizing the rpc's.
 	 */
-	biosize = nmp->nm_rsize;
+	biosize = nmp->nm_rpcclnt.rc_rsize;
 	do {
 
 		/*
@@ -494,11 +486,7 @@ nfs_getcacheblk(vp, bn, size, p)
 	if (nmp->nm_flag & NFSMNT_INT) {
 		bp = getblk(vp, bn, size, PCATCH, 0);
 		while (bp == (struct buf *)0) {
-			#ifdef RPCCLNT
-			if (nfs_sigintr(nmp, (struct rpctask *)NULL, p))
-			#else
-			if (nfs_sigintr(nmp, (struct nfsreq *)NULL, p))
-			#endif
+			if (nfs_sigintr(nmp, NULL, p))
 				return ((struct buf *)0);
 			bp = getblk(vp, bn, size, 0, 2 * hz);
 		}
@@ -539,11 +527,7 @@ nfs_vinvalbuf(vp, flags, cred, p, intrflg)
 		np->n_flag |= NFLUSHWANT;
 		error = tsleep((caddr_t)&np->n_flag, PRIBIO + 2, "nfsvinval",
 			slptimeo);
-		#ifdef RPCCLNT
-		if(error && intrflg && nfs_sigintr(nmp, (struct rpctask *)0, p))
-		#else
-		if (error && intrflg && nfs_sigintr(nmp, (struct nfsreq *)0, p))
-		#endif
+		if(error && intrflg && nfs_sigintr(nmp, NULL, p))
 			return (EINTR);
 	}
 
@@ -553,11 +537,7 @@ nfs_vinvalbuf(vp, flags, cred, p, intrflg)
 	np->n_flag |= NFLUSHINPROG;
 	error = vinvalbuf(vp, flags, cred, p, slpflag, 0);
 	while (error) {
-		#ifdef RPCCLNT
-		if (intrflg && nfs_sigintr(nmp, (struct rpctask *)0, p)) {
-		#else
-		if (intrflg && nfs_sigintr(nmp, (struct nfsreq *)0, p)) {
-		#endif
+		if (intrflg && nfs_sigintr(nmp, NULL, p)) {
 			np->n_flag &= ~NFLUSHINPROG;
 			if (np->n_flag & NFLUSHWANT) {
 				np->n_flag &= ~NFLUSHWANT;
@@ -589,16 +569,16 @@ nfs_asyncio(bp)
 	if (nfs_numasync == 0)
 		return (EIO);
 	for (i = 0; i < NFS_MAXASYNCDAEMON; i++)
-	    if (nfs_iodwant[i]) {
-		if ((bp->b_flags & B_READ) == 0) {
-			bp->b_flags |= B_WRITEINPROG;
-		}
+		if (nfs_iodwant[i]) {
+			if ((bp->b_flags & B_READ) == 0) {
+				bp->b_flags |= B_WRITEINPROG;
+			}
 	
-		TAILQ_INSERT_TAIL(&nfs_bufq, bp, b_freelist);
-		nfs_iodwant[i] = (struct proc *)0;
-		wakeup((caddr_t)&nfs_iodwant[i]);
-		return (0);
-	    }
+			TAILQ_INSERT_TAIL(&nfs_bufq, bp, b_freelist);
+			nfs_iodwant[i] = (struct proc *)0;
+			wakeup((caddr_t)&nfs_iodwant[i]);
+			return (0);
+		}
 
 	/*
 	 * If it is a read or a write already marked B_WRITEINPROG or B_NOCACHE
@@ -616,8 +596,8 @@ nfs_asyncio(bp)
 	 */
 	s = splbio();
 	buf_dirty(bp);
-	splx(s);
 	biodone(bp);
+	splx(s);
 	return (0);
 }
 
@@ -651,132 +631,134 @@ nfs_doio(bp, p)
 	 * Historically, paging was done with physio, but no more...
 	 */
 	if (bp->b_flags & B_PHYS) {
-	    /*
-	     * ...though reading /dev/drum still gets us here.
-	     */
-	    io.iov_len = uiop->uio_resid = bp->b_bcount;
-	    /* mapping was done by vmapbuf() */
-	    io.iov_base = bp->b_data;
-	    uiop->uio_offset = ((off_t)bp->b_blkno) << DEV_BSHIFT;
-	    if (bp->b_flags & B_READ) {
-		uiop->uio_rw = UIO_READ;
-		nfsstats.read_physios++;
-		error = nfs_readrpc(vp, uiop, curproc->p_ucred);
-	    } else {
-		iomode = NFSV3WRITE_DATASYNC;
-		uiop->uio_rw = UIO_WRITE;
-		nfsstats.write_physios++;
-		error = nfs_writerpc(vp, uiop, curproc->p_ucred, &iomode, &must_commit);
-	    }
-	    if (error) {
-		bp->b_flags |= B_ERROR;
-		bp->b_error = error;
-	    }
-	} else if (bp->b_flags & B_READ) {
-	    io.iov_len = uiop->uio_resid = bp->b_bcount;
-	    io.iov_base = bp->b_data;
-	    uiop->uio_rw = UIO_READ;
-	    switch (vp->v_type) {
-	    case VREG:
+		/*
+		 * ...though reading /dev/drum still gets us here.
+		 */
+		io.iov_len = uiop->uio_resid = bp->b_bcount;
+		/* mapping was done by vmapbuf() */
+		io.iov_base = bp->b_data;
 		uiop->uio_offset = ((off_t)bp->b_blkno) << DEV_BSHIFT;
-		nfsstats.read_bios++;
-		error = nfs_readrpc(vp, uiop, curproc->p_ucred);
-		if (!error) {
-		    bp->b_validoff = 0;
-		    if (uiop->uio_resid) {
-			/*
-			 * If len > 0, there is a hole in the file and
-			 * no writes after the hole have been pushed to
-			 * the server yet.
-			 * Just zero fill the rest of the valid area.
-			 */
-			diff = bp->b_bcount - uiop->uio_resid;
-			len = np->n_size - ((((off_t)bp->b_blkno) << DEV_BSHIFT)
-				+ diff);
-			if (len > 0) {
-			    len = min(len, uiop->uio_resid);
-			    bzero((char *)bp->b_data + diff, len);
-			    bp->b_validend = diff + len;
-			} else
-			    bp->b_validend = diff;
-		    } else
-			bp->b_validend = bp->b_bcount;
+		if (bp->b_flags & B_READ) {
+			uiop->uio_rw = UIO_READ;
+			nfsstats.read_physios++;
+			error = nfs_readrpc(vp, uiop);
+		} else {
+			iomode = NFSV3WRITE_DATASYNC;
+			uiop->uio_rw = UIO_WRITE;
+			nfsstats.write_physios++;
+			error = nfs_writerpc(vp, uiop, &iomode, &must_commit);
 		}
-		if (p && (vp->v_flag & VTEXT) &&
-		    (np->n_mtime != np->n_vattr.va_mtime.tv_sec)) {
-			uprintf("Process killed due to text file modification\n");
-			psignal(p, SIGKILL);
-			p->p_holdcnt++;
-		}
-		break;
-	    case VLNK:
-		uiop->uio_offset = (off_t)0;
-		nfsstats.readlink_bios++;
-		error = nfs_readlinkrpc(vp, uiop, curproc->p_ucred);
-		break;
-	    default:
-		printf("nfs_doio:  type %x unexpected\n",vp->v_type);
-		break;
-	    };
-	    if (error) {
-		bp->b_flags |= B_ERROR;
-		bp->b_error = error;
-	    }
-	} else {
-	    io.iov_len = uiop->uio_resid = bp->b_dirtyend
-		- bp->b_dirtyoff;
-	    uiop->uio_offset = ((off_t)bp->b_blkno) * DEV_BSIZE
-		+ bp->b_dirtyoff;
-	    io.iov_base = (char *)bp->b_data + bp->b_dirtyoff;
-	    uiop->uio_rw = UIO_WRITE;
-	    nfsstats.write_bios++;
-	    if ((bp->b_flags & (B_ASYNC | B_NEEDCOMMIT | B_NOCACHE)) == B_ASYNC)
-		iomode = NFSV3WRITE_UNSTABLE;
-	    else
-		iomode = NFSV3WRITE_FILESYNC;
-	    bp->b_flags |= B_WRITEINPROG;
-#ifdef fvdl_debug
-	    printf("nfs_doio(%x): bp %x doff %d dend %d\n", 
-		vp, bp, bp->b_dirtyoff, bp->b_dirtyend);
-#endif
-	    error = nfs_writerpc(vp, uiop, curproc->p_ucred, &iomode, &must_commit);
-	    if (!error && iomode == NFSV3WRITE_UNSTABLE)
-		bp->b_flags |= B_NEEDCOMMIT;
-	    else
-		bp->b_flags &= ~B_NEEDCOMMIT;
-	    bp->b_flags &= ~B_WRITEINPROG;
-
-	    /*
-	     * For an interrupted write, the buffer is still valid and the
-	     * write hasn't been pushed to the server yet, so we can't set
-	     * B_ERROR and report the interruption by setting B_EINTR. For
-	     * the B_ASYNC case, B_EINTR is not relevant, so the rpc attempt
-	     * is essentially a noop.
-	     * For the case of a V3 write rpc not being committed to stable
-	     * storage, the block is still dirty and requires either a commit
-	     * rpc or another write rpc with iomode == NFSV3WRITE_FILESYNC
-	     * before the block is reused. This is indicated by setting the
-	     * B_DELWRI and B_NEEDCOMMIT flags.
-	     */
-	    if (error == EINTR || (!error && (bp->b_flags & B_NEEDCOMMIT))) {
-		    s = splbio();
-		    buf_dirty(bp);
-		    splx(s);
-
-		    if (!(bp->b_flags & B_ASYNC) && error)
-			    bp->b_flags |= B_EINTR;
-	    } else {
 		if (error) {
-		    bp->b_flags |= B_ERROR;
-		    bp->b_error = np->n_error = error;
-		    np->n_flag |= NWRITEERR;
+			bp->b_flags |= B_ERROR;
+			bp->b_error = error;
+		}
+	} else if (bp->b_flags & B_READ) {
+		io.iov_len = uiop->uio_resid = bp->b_bcount;
+		io.iov_base = bp->b_data;
+		uiop->uio_rw = UIO_READ;
+		switch (vp->v_type) {
+		case VREG:
+			uiop->uio_offset = ((off_t)bp->b_blkno) << DEV_BSHIFT;
+			nfsstats.read_bios++;
+			error = nfs_readrpc(vp, uiop);
+			if (!error) {
+				bp->b_validoff = 0;
+				if (uiop->uio_resid) {
+					/*
+					 * If len > 0, there is a hole in the file and
+					 * no writes after the hole have been pushed to
+					 * the server yet.
+					 * Just zero fill the rest of the valid area.
+					 */
+					diff = bp->b_bcount - uiop->uio_resid;
+					len = np->n_size - ((((off_t)bp->b_blkno) << DEV_BSHIFT)
+							    + diff);
+					if (len > 0) {
+						len = min(len, uiop->uio_resid);
+						bzero((char *)bp->b_data + diff, len);
+						bp->b_validend = diff + len;
+					} else
+						bp->b_validend = diff;
+				} else
+					bp->b_validend = bp->b_bcount;
+			}
+			if (p && (vp->v_flag & VTEXT) &&
+			    (np->n_mtime != np->n_vattr.va_mtime.tv_sec)) {
+				uprintf("Process killed due to text file modification\n");
+				psignal(p, SIGKILL);
+				p->p_holdcnt++;
+			}
+			break;
+		case VLNK:
+			uiop->uio_offset = (off_t)0;
+			nfsstats.readlink_bios++;
+			error = nfs_readlinkrpc(vp, uiop, curproc->p_ucred);
+			break;
+		default:
+			printf("nfs_doio:  type %x unexpected\n",vp->v_type);
+			break;
+		};
+		if (error) {
+			bp->b_flags |= B_ERROR;
+			bp->b_error = error;
+		}
+	} else {
+		io.iov_len = uiop->uio_resid = bp->b_dirtyend
+			- bp->b_dirtyoff;
+		uiop->uio_offset = ((off_t)bp->b_blkno) * DEV_BSIZE
+			+ bp->b_dirtyoff;
+		io.iov_base = (char *)bp->b_data + bp->b_dirtyoff;
+		uiop->uio_rw = UIO_WRITE;
+		nfsstats.write_bios++;
+		if ((bp->b_flags & (B_ASYNC | B_NEEDCOMMIT | B_NOCACHE)) == B_ASYNC)
+			iomode = NFSV3WRITE_UNSTABLE;
+		else
+			iomode = NFSV3WRITE_FILESYNC;
+		bp->b_flags |= B_WRITEINPROG;
+#ifdef fvdl_debug
+		printf("nfs_doio(%x): bp %x doff %d dend %d\n", 
+		       vp, bp, bp->b_dirtyoff, bp->b_dirtyend);
+#endif
+		error = nfs_writerpc(vp, uiop, &iomode, &must_commit);
+		if (!error && iomode == NFSV3WRITE_UNSTABLE)
+			bp->b_flags |= B_NEEDCOMMIT;
+		else
+			bp->b_flags &= ~B_NEEDCOMMIT;
+		bp->b_flags &= ~B_WRITEINPROG;
+
+		/*
+		 * For an interrupted write, the buffer is still valid and the
+		 * write hasn't been pushed to the server yet, so we can't set
+		 * B_ERROR and report the interruption by setting B_EINTR. For
+		 * the B_ASYNC case, B_EINTR is not relevant, so the rpc attempt
+		 * is essentially a noop.
+		 * For the case of a V3 write rpc not being committed to stable
+		 * storage, the block is still dirty and requires either a commit
+		 * rpc or another write rpc with iomode == NFSV3WRITE_FILESYNC
+		 * before the block is reused. This is indicated by setting the
+		 * B_DELWRI and B_NEEDCOMMIT flags.
+		 */
+		if (error == EINTR || (!error && (bp->b_flags & B_NEEDCOMMIT))) {
+			s = splbio();
+			buf_dirty(bp);
+			splx(s);
+
+			if (!(bp->b_flags & B_ASYNC) && error)
+				bp->b_flags |= B_EINTR;
+		} else {
+			if (error) {
+				bp->b_flags |= B_ERROR;
+				bp->b_error = np->n_error = error;
+				np->n_flag |= NWRITEERR;
+			}
+			bp->b_dirtyoff = bp->b_dirtyend = 0;
 		}
-		bp->b_dirtyoff = bp->b_dirtyend = 0;
-	    }
 	}
 	bp->b_resid = uiop->uio_resid;
 	if (must_commit)
 		nfs_clearcommit(vp->v_mount);
+	s = splbio();
 	biodone(bp);
+	splx(s);
 	return (error);
 }
