@@ -1,4 +1,4 @@
-/*	$Id: nfs4_prot.h,v 1.11 2002/09/10 22:27:22 rees Exp $	*/
+/*	$Id: nfs4_prot.h,v 1.12 2003/06/20 17:05:09 rees Exp $	*/
 
 /*
  * copyright (c) 2002
@@ -31,6 +31,10 @@
 #ifndef _NFS4_PROT_H_RPCGEN
 #define _NFS4_PROT_H_RPCGEN
 
+#ifndef NFS_VER4
+#define NFS_VER4	4
+#endif
+
 #define xdr_uint32_t xdr_u_int32_t
 
 #ifndef XDR_QUADLEN
