@@ -1,4 +1,4 @@
-/*	$Id: nfs4_subs.c,v 1.15 2002/09/04 23:11:51 muzzle Exp $	*/
+/*	$Id: nfs4_subs.c,v 1.19 2003/06/20 17:05:10 rees Exp $	*/
 /*	$OpenBSD: nfs_subs.c,v 1.42 2002/03/14 01:27:13 millert Exp $	*/
 /*	$NetBSD: nfs_subs.c,v 1.27.4.3 1996/07/08 20:34:24 jtc Exp $	*/
 
@@ -71,8 +71,9 @@
  * copy data between mbuf chains and uio lists.
  */
 #include <sys/param.h>
-#include <sys/proc.h>
 #include <sys/systm.h>
+#include <sys/buf.h>
+#include <sys/proc.h>
 #include <sys/kernel.h>
 #include <sys/mount.h>
 #include <sys/vnode.h>
@@ -91,18 +92,16 @@
 #include <nfs/nfs.h>
 #include <nfs/xdr_subs.h>
 #include <nfs4/list.h>
-#include <nfs4/nfs4mount.h>
 #include <nfs4/nfs4_prot.h>
 #include <nfs4/nfs4_debug.h>
 #include <nfs4/nfs4.h>
+#include <nfs4/nfs4mount.h>
 #include <nfs4/nfs4fs.h>
 #include <nfs4/nfs4node.h>
 #include <nfs4/nfsm_subs.h>
 
 #include <nfs4/nfs4_var.h>
 
-#include <miscfs/specfs/specdev.h>
-
 #include <netinet/in.h>
 #ifdef ISO
 #include <netiso/iso.h>
@@ -140,6 +139,12 @@ nfstype nfsv3_type[9] = { NFNON, NFREG, NFDIR, NFBLK, NFCHR, NFLNK, NFSOCK,
 enum vtype nv2tov_type[8] = { VNON, VREG, VDIR, VBLK, VCHR, VLNK, VNON, VNON };
 enum vtype nv3tov_type[8]={ VNON, VREG, VDIR, VBLK, VCHR, VLNK, VSOCK, VFIFO };
 
+static uint32_t __fsinfo_bm[2], __fsattr_bm[2], __getattr_bm[2], __readdir_bm[2];
+bitmap4 fsinfo_bm = { 2, __fsinfo_bm };
+bitmap4 fsattr_bm = { 2, __fsattr_bm };
+bitmap4 getattr_bm = { 2, __getattr_bm };
+bitmap4 readdir_bm = { 2, __readdir_bm };
+
 /*
  * Mapping of old NFS Version 2 RPC numbers to generic numbers.
  */
@@ -562,12 +567,6 @@ static short *nfsrv_v3errmap[] = {
 };
 
 extern struct proc *nfs_iodwant[NFS_MAXASYNCDAEMON];
-
-extern struct rpcstats rpcstats;
-
-extern struct nfsstats nfsstats;
-extern nfstype nfsv2_type[9];
-extern nfstype nfsv3_type[9];
 extern struct nfsnodehashhead *nfsnodehashtbl;
 extern u_long nfsnodehash;
 
@@ -927,7 +926,7 @@ nfs_init()
 	nfs_true = txdr_unsigned(TRUE);
 	nfs_false = txdr_unsigned(FALSE);
 	nfs_xdrneg1 = txdr_unsigned(-1);
-	
+
 #ifdef NFSSERVER
 	nfsrv_init(0);			/* Init server data structures */
 	nfsrv_initcache();		/* Init the server request cache */
@@ -936,7 +935,45 @@ nfs_init()
 	rpcclnt_init();	
 }
 
-#ifdef NFSV4CLIENT
+void
+nfs4_setup_bitmaps()
+{
+	static int inited;
+
+	if (inited)
+		return;
+	inited = 1;
+
+	FATTR4_SET(FATTR4_FSID, __fsinfo_bm);
+	FATTR4_SET(FATTR4_MAXREAD, __fsinfo_bm);
+	FATTR4_SET(FATTR4_MAXWRITE, __fsinfo_bm);
+	FATTR4_SET(FATTR4_LEASE_TIME, __fsinfo_bm);
+
+	FATTR4_SET(FATTR4_FSID, __fsattr_bm);
+	FATTR4_SET(FATTR4_FILES_FREE, __fsattr_bm);
+	FATTR4_SET(FATTR4_FILES_TOTAL, __fsattr_bm);
+	FATTR4_SET(FATTR4_SPACE_AVAIL, __fsattr_bm);
+	FATTR4_SET(FATTR4_SPACE_FREE, __fsattr_bm);
+	FATTR4_SET(FATTR4_SPACE_TOTAL, __fsattr_bm);
+
+	FATTR4_SET(FATTR4_TYPE, __getattr_bm);
+	FATTR4_SET(FATTR4_FSID, __getattr_bm);
+	FATTR4_SET(FATTR4_SIZE, __getattr_bm);
+	FATTR4_SET(FATTR4_MODE, __getattr_bm);
+	FATTR4_SET(FATTR4_RAWDEV, __getattr_bm);
+	FATTR4_SET(FATTR4_NUMLINKS, __getattr_bm);
+	FATTR4_SET(FATTR4_OWNER, __getattr_bm);
+	FATTR4_SET(FATTR4_OWNER_GROUP, __getattr_bm);
+	FATTR4_SET(FATTR4_FILEID, __getattr_bm);
+	FATTR4_SET(FATTR4_TIME_MODIFY, __getattr_bm);
+	FATTR4_SET(FATTR4_TIME_ACCESS, __getattr_bm);
+	FATTR4_SET(FATTR4_TIME_CREATE, __getattr_bm);
+
+	FATTR4_SET(FATTR4_TYPE, __readdir_bm);
+	FATTR4_SET(FATTR4_FSID, __readdir_bm);
+	FATTR4_SET(FATTR4_FILEID, __readdir_bm);
+}
+
 int
 nfs_vfs_init(vfsp)
 	struct vfsconf *vfsp;
@@ -991,14 +1028,14 @@ nfs4_loadattrcache(struct vnode *vp, struct nfs4fs_fattr *fp, struct vattr *vape
 	 */
 	np = VTONFS(vp);
 	vap = &np->n_vattr;
-	if (vp->v_type != vtyp) {
+	if (vp->v_type != vtyp || np->n_mtime == 0) {
 		bzero(vap, sizeof *vap);
 		vp->v_type = vtyp;
 		np->n_mtime = mtime.tv_sec;
 	}
 	vap->va_type = vtyp;
 	vap->va_mode = (vmode & 07777);
-	vap->va_rdev = (dev_t)rdev;
+	vap->va_rdev = rdev;
 	vap->va_mtime = mtime;
 	vap->va_fsid = vp->v_mount->mnt_stat.f_fsid.val[0];
 	if (fp->fa_valid & FA_NLINK)
@@ -1111,403 +1148,6 @@ nfs_getattrcache(vp, vaper)
 	}
 	return (0);
 }
-#endif /* NFSV4CLIENT */
-
-/*
- * Set up nameidata for a lookup() call and do it
- */
-int
-nfs_namei(ndp, fhp, len, slp, nam, mdp, dposp, retdirp, p, kerbflag)
-	struct nameidata *ndp;
-	fhandle_t *fhp;
-	int len;
-	struct nfssvc_sock *slp;
-	struct mbuf *nam;
-	struct mbuf **mdp;
-	caddr_t *dposp;
-	struct vnode **retdirp;
-	struct proc *p;
-	int kerbflag;
-{
-	int i, rem;
-	struct mbuf *md;
-	char *fromcp, *tocp;
-	struct vnode *dp;
-	int error, rdonly;
-	struct componentname *cnp = &ndp->ni_cnd;
-
-	*retdirp = (struct vnode *)0;
-	MALLOC(cnp->cn_pnbuf, char *, len + 1, M_NAMEI, M_WAITOK);
-	/*
-	 * Copy the name from the mbuf list to ndp->ni_pnbuf
-	 * and set the various ndp fields appropriately.
-	 */
-	fromcp = *dposp;
-	tocp = cnp->cn_pnbuf;
-	md = *mdp;
-	rem = mtod(md, caddr_t) + md->m_len - fromcp;
-	cnp->cn_hash = 0;
-	for (i = 0; i < len; i++) {
-		while (rem == 0) {
-			md = md->m_next;
-			if (md == NULL) {
-				error = EBADRPC;
-				goto out;
-			}
-			fromcp = mtod(md, caddr_t);
-			rem = md->m_len;
-		}
-		if (*fromcp == '\0' || *fromcp == '/') {
-			error = EACCES;
-			goto out;
-		}
-		cnp->cn_hash += (u_char)*fromcp;
-		*tocp++ = *fromcp++;
-		rem--;
-	}
-	*tocp = '\0';
-	*mdp = md;
-	*dposp = fromcp;
-	len = nfsm_rndup(len)-len;
-	if (len > 0) {
-		if (rem >= len)
-			*dposp += len;
-		else if ((error = nfs_adv(mdp, dposp, len, rem)) != 0)
-			goto out;
-	}
-	ndp->ni_pathlen = tocp - cnp->cn_pnbuf;
-	cnp->cn_nameptr = cnp->cn_pnbuf;
-	/*
-	 * Extract and set starting directory.
-	 */
-	error = nfsrv_fhtovp(fhp, FALSE, &dp, ndp->ni_cnd.cn_cred, slp,
-	    nam, &rdonly, kerbflag);
-	if (error)
-		goto out;
-	if (dp->v_type != VDIR) {
-		vrele(dp);
-		error = ENOTDIR;
-		goto out;
-	}
-	VREF(dp);
-	*retdirp = dp;
-	ndp->ni_startdir = dp;
-	if (rdonly)
-		cnp->cn_flags |= (NOCROSSMOUNT | RDONLY);
-	else
-		cnp->cn_flags |= NOCROSSMOUNT;
-	/*
-	 * And call lookup() to do the real work
-	 */
-	cnp->cn_proc = p;
-	error = lookup(ndp);
-	if (error)
-		goto out;
-	/*
-	 * Check for encountering a symbolic link
-	 */
-	if (cnp->cn_flags & ISSYMLINK) {
-		if ((cnp->cn_flags & LOCKPARENT) && ndp->ni_pathlen == 1)
-			vput(ndp->ni_dvp);
-		else
-			vrele(ndp->ni_dvp);
-		vput(ndp->ni_vp);
-		ndp->ni_vp = NULL;
-		error = EINVAL;
-		goto out;
-	}
-	/*
-	 * Check for saved name request
-	 */
-	if (cnp->cn_flags & (SAVENAME | SAVESTART)) {
-		cnp->cn_flags |= HASBUF;
-		return (0);
-	}
-out:
-	FREE(cnp->cn_pnbuf, M_NAMEI);
-	return (error);
-}
-
-/*
- * A fiddled version of m_adj() that ensures null fill to a long
- * boundary and only trims off the back end
- */
-void
-nfsm_adj(mp, len, nul)
-	struct mbuf *mp;
-	int len;
-	int nul;
-{
-	struct mbuf *m;
-	int count, i;
-	char *cp;
-
-	/*
-	 * Trim from tail.  Scan the mbuf chain,
-	 * calculating its length and finding the last mbuf.
-	 * If the adjustment only affects this mbuf, then just
-	 * adjust and return.  Otherwise, rescan and truncate
-	 * after the remaining size.
-	 */
-	count = 0;
-	m = mp;
-	for (;;) {
-		count += m->m_len;
-		if (m->m_next == (struct mbuf *)0)
-			break;
-		m = m->m_next;
-	}
-	if (m->m_len > len) {
-		m->m_len -= len;
-		if (nul > 0) {
-			cp = mtod(m, caddr_t)+m->m_len-nul;
-			for (i = 0; i < nul; i++)
-				*cp++ = '\0';
-		}
-		return;
-	}
-	count -= len;
-	if (count < 0)
-		count = 0;
-	/*
-	 * Correct length for chain is "count".
-	 * Find the mbuf with last data, adjust its length,
-	 * and toss data from remaining mbufs on chain.
-	 */
-	for (m = mp; m; m = m->m_next) {
-		if (m->m_len >= count) {
-			m->m_len = count;
-			if (nul > 0) {
-				cp = mtod(m, caddr_t)+m->m_len-nul;
-				for (i = 0; i < nul; i++)
-					*cp++ = '\0';
-			}
-			break;
-		}
-		count -= m->m_len;
-	}
-	for (m = m->m_next;m;m = m->m_next)
-		m->m_len = 0;
-}
-
-/*
- * Make these functions instead of macros, so that the kernel text size
- * doesn't get too big...
- */
-void
-nfsm_srvwcc(nfsd, before_ret, before_vap, after_ret, after_vap, mbp, bposp)
-	struct nfsrv_descript *nfsd;
-	int before_ret;
-	struct vattr *before_vap;
-	int after_ret;
-	struct vattr *after_vap;
-	struct mbuf **mbp;
-	char **bposp;
-{
-	struct mbuf *mb = *mbp, *mb2;
-	char *bpos = *bposp;
-	u_int32_t *tl;
-
-	if (before_ret) {
-		nfsm_build(tl, u_int32_t *, NFSX_UNSIGNED);
-		*tl = nfs_false;
-	} else {
-		nfsm_build(tl, u_int32_t *, 7 * NFSX_UNSIGNED);
-		*tl++ = nfs_true;
-		txdr_hyper(before_vap->va_size, tl);
-		tl += 2;
-		txdr_nfsv3time(&(before_vap->va_mtime), tl);
-		tl += 2;
-		txdr_nfsv3time(&(before_vap->va_ctime), tl);
-	}
-	*bposp = bpos;
-	*mbp = mb;
-	nfsm_srvpostopattr(nfsd, after_ret, after_vap, mbp, bposp);
-}
-
-void
-nfsm_srvpostopattr(nfsd, after_ret, after_vap, mbp, bposp)
-	struct nfsrv_descript *nfsd;
-	int after_ret;
-	struct vattr *after_vap;
-	struct mbuf **mbp;
-	char **bposp;
-{
-	struct mbuf *mb = *mbp, *mb2;
-	char *bpos = *bposp;
-	u_int32_t *tl;
-	struct nfs_fattr *fp;
-
-	if (after_ret) {
-		nfsm_build(tl, u_int32_t *, NFSX_UNSIGNED);
-		*tl = nfs_false;
-	} else {
-		nfsm_build(tl, u_int32_t *, NFSX_UNSIGNED + NFSX_V3FATTR);
-		*tl++ = nfs_true;
-		fp = (struct nfs_fattr *)tl;
-		nfsm_srvfattr(nfsd, after_vap, fp);
-	}
-	*mbp = mb;
-	*bposp = bpos;
-}
-
-void
-nfsm_srvfattr(nfsd, vap, fp)
-	struct nfsrv_descript *nfsd;
-	struct vattr *vap;
-	struct nfs_fattr *fp;
-{
-
-	fp->fa_nlink = txdr_unsigned(vap->va_nlink);
-	fp->fa_uid = txdr_unsigned(vap->va_uid);
-	fp->fa_gid = txdr_unsigned(vap->va_gid);
-	if (nfsd->nd_flag & ND_NFSV3) {
-		fp->fa_type = vtonfsv3_type(vap->va_type);
-		fp->fa_mode = vtonfsv3_mode(vap->va_mode);
-		txdr_hyper(vap->va_size, &fp->fa3_size);
-		txdr_hyper(vap->va_bytes, &fp->fa3_used);
-		fp->fa3_rdev.specdata1 = txdr_unsigned(major(vap->va_rdev));
-		fp->fa3_rdev.specdata2 = txdr_unsigned(minor(vap->va_rdev));
-		fp->fa3_fsid.nfsuquad[0] = 0;
-		fp->fa3_fsid.nfsuquad[1] = txdr_unsigned(vap->va_fsid);
-		fp->fa3_fileid.nfsuquad[0] = 0;
-		fp->fa3_fileid.nfsuquad[1] = txdr_unsigned(vap->va_fileid);
-		txdr_nfsv3time(&vap->va_atime, &fp->fa3_atime);
-		txdr_nfsv3time(&vap->va_mtime, &fp->fa3_mtime);
-		txdr_nfsv3time(&vap->va_ctime, &fp->fa3_ctime);
-	} else {
-		fp->fa_type = vtonfsv2_type(vap->va_type);
-		fp->fa_mode = vtonfsv2_mode(vap->va_type, vap->va_mode);
-		fp->fa2_size = txdr_unsigned(vap->va_size);
-		fp->fa2_blocksize = txdr_unsigned(vap->va_blocksize);
-		if (vap->va_type == VFIFO)
-			fp->fa2_rdev = 0xffffffff;
-		else
-			fp->fa2_rdev = txdr_unsigned(vap->va_rdev);
-		fp->fa2_blocks = txdr_unsigned(vap->va_bytes / NFS_FABLKSIZE);
-		fp->fa2_fsid = txdr_unsigned(vap->va_fsid);
-		fp->fa2_fileid = txdr_unsigned(vap->va_fileid);
-		txdr_nfsv2time(&vap->va_atime, &fp->fa2_atime);
-		txdr_nfsv2time(&vap->va_mtime, &fp->fa2_mtime);
-		txdr_nfsv2time(&vap->va_ctime, &fp->fa2_ctime);
-	}
-}
-
-/*
- * nfsrv_fhtovp() - convert a fh to a vnode ptr (optionally locked)
- * 	- look up fsid in mount list (if not found ret error)
- *	- get vp and export rights by calling VFS_FHTOVP() and VFS_CHECKEXP()
- *	- if cred->cr_uid == 0 or MNT_EXPORTANON set it to credanon
- *	- if not lockflag unlock it with VOP_UNLOCK()
- */
-int
-nfsrv_fhtovp(fhp, lockflag, vpp, cred, slp, nam, rdonlyp, kerbflag)
-	fhandle_t *fhp;
-	int lockflag;
-	struct vnode **vpp;
-	struct ucred *cred;
-	struct nfssvc_sock *slp;
-	struct mbuf *nam;
-	int *rdonlyp;
-	int kerbflag;
-{
-	struct proc *p = curproc;	/* XXX */
-	struct mount *mp;
-	int i;
-	struct ucred *credanon;
-	int error, exflags;
-	struct sockaddr_in *saddr;
-
-	*vpp = (struct vnode *)0;
-	mp = vfs_getvfs(&fhp->fh_fsid);
-
-	if (!mp)
-		return (ESTALE);
-	error = VFS_CHECKEXP(mp, nam, &exflags, &credanon);
-	if (error)
-		return (error);
-	error = VFS_FHTOVP(mp, &fhp->fh_fid, vpp);
-	if (error)
-		return (error);
-
-	saddr = mtod(nam, struct sockaddr_in *);
-	if (saddr->sin_family == AF_INET &&
-	    (ntohs(saddr->sin_port) >= IPPORT_RESERVED ||
-	    (slp->ns_so->so_type == SOCK_STREAM && ntohs(saddr->sin_port) == 20))) {
-		vput(*vpp);
-		return (NFSERR_AUTHERR | AUTH_TOOWEAK);
-	}
-
-	/*
-	 * Check/setup credentials.
-	 */
-	if (exflags & MNT_EXKERB) {
-		if (!kerbflag) {
-			vput(*vpp);
-			return (NFSERR_AUTHERR | AUTH_TOOWEAK);
-		}
-	} else if (kerbflag) {
-		vput(*vpp);
-		return (NFSERR_AUTHERR | AUTH_TOOWEAK);
-	} else if (cred->cr_uid == 0 || (exflags & MNT_EXPORTANON)) {
-		cred->cr_uid = credanon->cr_uid;
-		cred->cr_gid = credanon->cr_gid;
-		for (i = 0; i < credanon->cr_ngroups && i < NGROUPS; i++)
-			cred->cr_groups[i] = credanon->cr_groups[i];
-		cred->cr_ngroups = i;
-	}
-	if (exflags & MNT_EXRDONLY)
-		*rdonlyp = 1;
-	else
-		*rdonlyp = 0;
-	if (!lockflag)
-		VOP_UNLOCK(*vpp, 0, p);
-
-	return (0);
-}
-
-/*
- * This function compares two net addresses by family and returns TRUE
- * if they are the same host.
- * If there is any doubt, return FALSE.
- * The AF_INET family is handled as a special case so that address mbufs
- * don't need to be saved to store "struct in_addr", which is only 4 bytes.
- */
-int
-netaddr_match(family, haddr, nam)
-	int family;
-	union nethostaddr *haddr;
-	struct mbuf *nam;
-{
-	struct sockaddr_in *inetaddr;
-
-	switch (family) {
-	case AF_INET:
-		inetaddr = mtod(nam, struct sockaddr_in *);
-		if (inetaddr->sin_family == AF_INET &&
-		    inetaddr->sin_addr.s_addr == haddr->had_inetaddr)
-			return (1);
-		break;
-#ifdef ISO
-	case AF_ISO:
-	    {
-		struct sockaddr_iso *isoaddr1, *isoaddr2;
-
-		isoaddr1 = mtod(nam, struct sockaddr_iso *);
-		isoaddr2 = mtod(haddr->had_nam, struct sockaddr_iso *);
-		if (isoaddr1->siso_family == AF_ISO &&
-		    isoaddr1->siso_nlen > 0 &&
-		    isoaddr1->siso_nlen == isoaddr2->siso_nlen &&
-		    SAME_ISOADDR(isoaddr1, isoaddr2))
-			return (1);
-		break;
-	    }
-#endif	/* ISO */
-	default:
-		break;
-	};
-	return (0);
-}
 
 /*
  * The write verifier has changed (probably due to a server reboot), so all
@@ -1516,14 +1156,14 @@ netaddr_match(family, haddr, nam)
  * flag. Once done the new write verifier can be set for the mount point.
  */
 void
-nfs_clearcommit(mp)
-	struct mount *mp;
+nfs_clearcommit(struct mount *mp)
 {
 	struct vnode *vp, *nvp;
 	struct buf *bp, *nbp;
 	int s;
 
 	s = splbio();
+#if defined(__OpenBSD__)
 loop:
 	for (vp = LIST_FIRST(&mp->mnt_vnodelist); vp != NULL; vp = nvp) {
 		if (vp->v_mount != mp)	/* Paranoia */
@@ -1536,6 +1176,27 @@ loop:
 				bp->b_flags &= ~B_NEEDCOMMIT;
 		}
 	}
+#elif defined(__FreeBSD__)
+	mtx_lock(&mntvnode_mtx);
+loop:
+	for (vp = TAILQ_FIRST(&mp->mnt_nvnodelist); vp; vp = nvp) {
+		if (vp->v_mount != mp)	/* Paranoia */
+			goto loop;
+		nvp = TAILQ_NEXT(vp, v_nmntvnodes);
+		VI_LOCK(vp);
+		mtx_unlock(&mntvnode_mtx);
+		for (bp = TAILQ_FIRST(&vp->v_dirtyblkhd); bp; bp = nbp) {
+			nbp = TAILQ_NEXT(bp, b_vnbufs);
+			if (BUF_REFCNT(bp) == 0 &&
+			    (bp->b_flags & (B_DELWRI | B_NEEDCOMMIT))
+				== (B_DELWRI | B_NEEDCOMMIT))
+				bp->b_flags &= ~(B_NEEDCOMMIT | B_CLUSTEROK);
+		}
+		VI_UNLOCK(vp);
+		mtx_lock(&mntvnode_mtx);
+	}
+	mtx_unlock(&mntvnode_mtx);
+#endif
 	splx(s);
 }
 
