@@ -1,4 +1,4 @@
-/*	$Id: nfs4_vfsops.c,v 1.20 2002/08/20 21:49:27 rees Exp $	*/
+/*	$Id: nfs4_vfsops.c,v 1.36 2003/06/20 17:05:11 rees Exp $	*/
 /*	$OpenBSD: nfs_vfsops.c,v 1.48 2002/03/14 01:27:13 millert Exp $	*/
 /*	$NetBSD: nfs_vfsops.c,v 1.46.4.1 1996/05/25 22:40:35 fvdl Exp $	*/
 
@@ -66,7 +66,6 @@
 
 #include <sys/param.h>
 #include <sys/conf.h>
-#include <sys/ioctl.h>
 #include <sys/signal.h>
 #include <sys/proc.h>
 #include <sys/namei.h>
@@ -79,43 +78,39 @@
 #include <sys/socketvar.h>
 #include <sys/systm.h>
 #include <sys/sysctl.h>
+#include <sys/kthread.h>
 
 #include <net/if.h>
 #include <net/route.h>
 #include <netinet/in.h>
+#include <netinet/in_var.h>
 
 #include <nfs/rpcv2.h>
 #include <nfs/nfsproto.h>
 #include <nfs/nfs.h>
 #include <nfs/xdr_subs.h>
-#include <nfs/nfsdiskless.h>
 #include <nfs4/list.h>
-#include <nfs4/nfs4_var.h>
-#include <nfs4/nfs4mount.h>
 #include <nfs4/nfs4_prot.h>
 #include <nfs4/nfs4_debug.h>
 #include <nfs4/nfs4.h>
+#include <nfs4/nfs4mount.h>
 #include <nfs4/nfs4fs.h>
 #include <nfs4/nfs4node.h>
 #include <nfs4/nfsm_subs.h>
+#include <nfs4/nfs4_var.h>
 
-#ifdef RPCCLNT
 #include <rpc/rpcclnt.h>
-#endif
 
-extern struct nfsstats nfsstats;
-
-#ifdef RPCCLNT
 extern int rpcclnt_ticks;
-#else
-extern int nfs_ticks;
-#endif
 
-int mountnfs(struct nfs_args *argp, struct mount *mp,
-	     struct mbuf *nam, char *pth, char *hst, struct proc *p);
 int nfs_sysctl(int *, u_int, void *, size_t *, void *, size_t, struct proc *);
 int nfs_checkexp(struct mount *mp, struct mbuf *nam,
 	 int *extflagsp, struct ucred **credanonp);
+static int nfs4_do_renew(struct nfs4mount *nmp, struct ucred *cred);
+void nfs4_daemon(void *arg);
+int mountnfs(struct nfs_args *argp, struct mount *mp,
+	     struct mbuf *nam, char *pth, char *hst, struct proc *p);
+static int nfs4_stuff_fsattr(struct nfs4fs_fattr *fap, struct nfs4mount *nmp, struct statfs *sbp);
 
 /*
  * nfs vfs operations.
@@ -136,20 +131,8 @@ struct vfsops nfs_vfsops = {
 	nfs_checkexp
 };
 
-extern u_int32_t nfs_procids[NFS_NPROCS];
 extern u_int32_t nfs_prog, nfs_vers;
 
-struct mount *nfs_mount_diskless
-    __P((struct nfs_dlmount *, char *, int, struct vnode **));
-
-static int nfs4_stuff_fsattr(struct nfs4fs_fattr *fap, struct nfs4mount *nmp, struct statfs *sbp);
-
-static uint32_t __fsattr_bm[] = { 0xc0c00000, 0x00001c00 };
-bitmap4 fsattr_bm = { 2, __fsattr_bm };
-extern bitmap4 getattr_bm;
-
-char *strtok(char *s, const char *delim);
-
 #define TRUE	1
 #define	FALSE	0
 
@@ -184,7 +167,7 @@ nfs_statfs(mp, sbp, p)
 	compound.procp = p;
 	compound.cred = cred;
 	getattr.flags = FL_NOREVAL;
-	getattr.bm = &getattr_bm;
+	getattr.bm = &fsattr_bm;
 	putfh.dentry = &np->n_dentry;
 
 	if ((error = nfs4_setup_compound(&compound, "nfs4_statfs()")))
@@ -201,10 +184,6 @@ nfs_statfs(mp, sbp, p)
 	if ((error = nfs4_handle_getattr(&compound, &getattr)))
 		goto nfsmout;
 
-	/* Parse the getattr results twice, once for the vnode and once for the fs */
-	error = nfs4_loadattrcache(vp, &getattr.fattr, NULL);
-	if (error)
-		goto nfsmout;
 	nfs4_stuff_fsattr(&getattr.fattr, nmp, sbp);
 
 	if (sbp != &mp->mnt_stat) {
@@ -226,7 +205,28 @@ nfs_statfs(mp, sbp, p)
 	return (error);
 }
 
-/* v4 version doesn't do the rpc, just makes sure fsinfo is there */
+static int
+nfs4_stuff_fsattr(struct nfs4fs_fattr *fp, struct nfs4mount *nmp, struct statfs *sbp)
+{
+	sbp->f_flags = nmp->nm_flag;
+	sbp->f_bsize = NFS_FABLKSIZE;
+
+	if (fp->fa_valid & FA_FSID) {
+		sbp->f_fsid.val[0] = fp->fa_fsid_major;
+		sbp->f_fsid.val[1] = fp->fa_fsid_minor;
+	}
+
+	sbp->f_ffree = (fp->fa_valid & FA_FFREE) ? fp->fa_ffree : 0;
+	sbp->f_files = (fp->fa_valid & FA_FTOTAL) ? fp->fa_ftotal : 0;
+
+	sbp->f_bavail = (fp->fa_valid & FA_SAVAIL) ? (fp->fa_savail / NFS_FABLKSIZE) : 500000;
+	sbp->f_bfree = (fp->fa_valid & FA_SFREE) ? (fp->fa_sfree / NFS_FABLKSIZE) : 500000;
+	sbp->f_blocks = (fp->fa_valid & FA_STOTAL) ? (fp->fa_stotal / NFS_FABLKSIZE) : 1000000;
+
+	return 0;
+}
+
+/* We got fsinfo at mount time, so nothing to do here */
 int
 nfs_fsinfo(nmp, vp, cred, p)
 	struct nfs4mount *nmp;
@@ -234,10 +234,6 @@ nfs_fsinfo(nmp, vp, cred, p)
 	struct ucred *cred;
 	struct proc *p;
 {
-	if (!(nmp->nm_flag & NFSMNT_GOTFSINFO)) {
-		printf("nfs_fsinfo: no info\n");
-		return EIO;
-	}
 	return 0;
 }
 
@@ -275,54 +271,48 @@ nfs_decode_args(nmp, argp, nargp)
 	splx(s);
 
 	if ((argp->flags & NFSMNT_TIMEO) && argp->timeo > 0) {
-		nmp->nm_timeo = (argp->timeo * NFS_HZ + 5) / 10;
-		if (nmp->nm_timeo < NFS_MINTIMEO)
-			nmp->nm_timeo = NFS_MINTIMEO;
-		else if (nmp->nm_timeo > NFS_MAXTIMEO)
-			nmp->nm_timeo = NFS_MAXTIMEO;
+		nmp->nm_rpcclnt.rc_timeo = (argp->timeo * NFS_HZ + 5) / 10;
+		if (nmp->nm_rpcclnt.rc_timeo < NFS_MINTIMEO)
+			nmp->nm_rpcclnt.rc_timeo = NFS_MINTIMEO;
+		else if (nmp->nm_rpcclnt.rc_timeo > NFS_MAXTIMEO)
+			nmp->nm_rpcclnt.rc_timeo = NFS_MAXTIMEO;
 	}
 
 	if ((argp->flags & NFSMNT_RETRANS) && argp->retrans > 1) {
-		nmp->nm_retry = argp->retrans;
-		if (nmp->nm_retry > NFS_MAXREXMIT)
-			nmp->nm_retry = NFS_MAXREXMIT;
+		nmp->nm_rpcclnt.rc_retry = argp->retrans;
+		if (nmp->nm_rpcclnt.rc_retry > NFS_MAXREXMIT)
+			nmp->nm_rpcclnt.rc_retry = NFS_MAXREXMIT;
 	}
 
-	if (argp->flags & NFSMNT_NFSV3) {
-		if (argp->sotype == SOCK_DGRAM)
-			maxio = NFS_MAXDGRAMDATA;
-		else
-			maxio = NFS_MAXDATA;
-	} else
-		maxio = NFS_V2MAXDATA;
+	maxio = (argp->sotype == SOCK_DGRAM) ? NFS_MAXDGRAMDATA : NFS_MAXDATA;
 
 	if ((argp->flags & NFSMNT_WSIZE) && argp->wsize > 0) {
-		int osize = nmp->nm_wsize;
-		nmp->nm_wsize = argp->wsize;
+		int osize = nmp->nm_rpcclnt.rc_wsize;
+		nmp->nm_rpcclnt.rc_wsize = argp->wsize;
 		/* Round down to multiple of blocksize */
-		nmp->nm_wsize &= ~(NFS_FABLKSIZE - 1);
-		if (nmp->nm_wsize <= 0)
-			nmp->nm_wsize = NFS_FABLKSIZE;
-		adjsock |= (nmp->nm_wsize != osize);
+		nmp->nm_rpcclnt.rc_wsize &= ~(NFS_FABLKSIZE - 1);
+		if (nmp->nm_rpcclnt.rc_wsize <= 0)
+			nmp->nm_rpcclnt.rc_wsize = NFS_FABLKSIZE;
+		adjsock |= (nmp->nm_rpcclnt.rc_wsize != osize);
 	}
-	if (nmp->nm_wsize > maxio)
-		nmp->nm_wsize = maxio;
-	if (nmp->nm_wsize > MAXBSIZE)
-		nmp->nm_wsize = MAXBSIZE;
+	if (nmp->nm_rpcclnt.rc_wsize > maxio)
+		nmp->nm_rpcclnt.rc_wsize = maxio;
+	if (nmp->nm_rpcclnt.rc_wsize > MAXBSIZE)
+		nmp->nm_rpcclnt.rc_wsize = MAXBSIZE;
 
 	if ((argp->flags & NFSMNT_RSIZE) && argp->rsize > 0) {
-		int osize = nmp->nm_rsize;
-		nmp->nm_rsize = argp->rsize;
+		int osize = nmp->nm_rpcclnt.rc_rsize;
+		nmp->nm_rpcclnt.rc_rsize = argp->rsize;
 		/* Round down to multiple of blocksize */
-		nmp->nm_rsize &= ~(NFS_FABLKSIZE - 1);
-		if (nmp->nm_rsize <= 0)
-			nmp->nm_rsize = NFS_FABLKSIZE;
-		adjsock |= (nmp->nm_rsize != osize);
+		nmp->nm_rpcclnt.rc_rsize &= ~(NFS_FABLKSIZE - 1);
+		if (nmp->nm_rpcclnt.rc_rsize <= 0)
+			nmp->nm_rpcclnt.rc_rsize = NFS_FABLKSIZE;
+		adjsock |= (nmp->nm_rpcclnt.rc_rsize != osize);
 	}
-	if (nmp->nm_rsize > maxio)
-		nmp->nm_rsize = maxio;
-	if (nmp->nm_rsize > MAXBSIZE)
-		nmp->nm_rsize = MAXBSIZE;
+	if (nmp->nm_rpcclnt.rc_rsize > maxio)
+		nmp->nm_rpcclnt.rc_rsize = maxio;
+	if (nmp->nm_rpcclnt.rc_rsize > MAXBSIZE)
+		nmp->nm_rpcclnt.rc_rsize = MAXBSIZE;
 
 	if ((argp->flags & NFSMNT_READDIRSIZE) && argp->readdirsize > 0) {
 		nmp->nm_readdirsize = argp->readdirsize;
@@ -331,7 +321,7 @@ nfs_decode_args(nmp, argp, nargp)
 		if (nmp->nm_readdirsize < NFS_DIRBLKSIZ)
 			nmp->nm_readdirsize = NFS_DIRBLKSIZ;
 	} else if (argp->flags & NFSMNT_RSIZE)
-		nmp->nm_readdirsize = nmp->nm_rsize;
+		nmp->nm_readdirsize = nmp->nm_rpcclnt.rc_rsize;
 
 	if (nmp->nm_readdirsize > maxio)
 		nmp->nm_readdirsize = maxio;
@@ -373,9 +363,13 @@ nfs_decode_args(nmp, argp, nargp)
 	if (nmp->nm_acdirmin > nmp->nm_acdirmax)
 	  nmp->nm_acdirmin = nmp->nm_acdirmax;
 
-	if (nmp->nm_so && adjsock) {
+	/* XXX: set deadthresh */
+	nmp->nm_rpcclnt.rc_deadthresh = argp->deadthresh;
+
+
+	if (nmp->nm_rpcclnt.rc_so && adjsock) {
 		nfs_disconnect(nmp);
-		if (nmp->nm_sotype == SOCK_DGRAM)
+		if (nmp->nm_rpcclnt.rc_sotype == SOCK_DGRAM)
 			while (nfs_connect(nmp, NULL)) {
 				printf("nfs_args: retrying connect\n");
 				(void) tsleep((caddr_t)&lbolt,
@@ -384,17 +378,18 @@ nfs_decode_args(nmp, argp, nargp)
 	}
 
 	/* Update nargp based on nmp */
-	nargp->wsize = nmp->nm_wsize;
-	nargp->rsize = nmp->nm_rsize;
+	nargp->wsize = nmp->nm_rpcclnt.rc_wsize;
+	nargp->rsize = nmp->nm_rpcclnt.rc_rsize;
 	nargp->readdirsize = nmp->nm_readdirsize;
-	nargp->timeo = nmp->nm_timeo;
-	nargp->retrans = nmp->nm_retry;
+	nargp->timeo = nmp->nm_rpcclnt.rc_timeo;
+	nargp->retrans = nmp->nm_rpcclnt.rc_retry;
 	nargp->maxgrouplist = nmp->nm_numgrps;
 	nargp->readahead = nmp->nm_readahead;
 	nargp->acregmin = nmp->nm_acregmin;
 	nargp->acregmax = nmp->nm_acregmax;
 	nargp->acdirmin = nmp->nm_acdirmin;
 	nargp->acdirmax = nmp->nm_acdirmax;
+	nargp->deadthresh = nmp->nm_rpcclnt.rc_deadthresh;
 }
 
 /*
@@ -416,9 +411,8 @@ nfs_mount(mp, path, data, ndp, p)
 	struct mbuf *nam;
 	char pth[MNAMELEN], hst[MNAMELEN];
 	size_t len;
-	u_char nfh[NFSX_V3FHMAX];
 
-	error = copyin (data, (caddr_t)&args, sizeof (args.version));
+	error = copyin (data, &args, sizeof (args.version));
 	if (error)
 		return (error);
 	if (args.version == 3) {
@@ -445,18 +439,9 @@ nfs_mount(mp, path, data, ndp, p)
 
 		if (nmp == NULL)
 			return (EIO);
-		/*
-		 * When doing an update, we can't change from or to
-		 * v3 and/or nqnfs.
-		 */
-		args.flags = (args.flags & ~(NFSMNT_NFSV3|NFSMNT_NQNFS)) |
-		    (nmp->nm_flag & (NFSMNT_NFSV3|NFSMNT_NQNFS));
 		nfs_decode_args(nmp, &args, &mp->mnt_stat.mount_info.nfs_args);
 		return (0);
 	}
-	error = copyin((caddr_t)args.fh, (caddr_t)nfh, args.fhsize);
-	if (error)
-		return (error);
 	error = copyinstr(path, pth, MNAMELEN-1, &len);
 	if (error)
 		return (error);
@@ -469,11 +454,67 @@ nfs_mount(mp, path, data, ndp, p)
 	error = sockargs(&nam, (caddr_t)args.addr, args.addrlen, MT_SONAME);
 	if (error)
 		return (error);
-	args.fh = nfh;
 	error = mountnfs(&args, mp, nam, pth, hst, p);
 	return (error);
 }
 
+/*
+ * renew should be done async
+ * should re-scan mount queue each time
+ */
+struct proc *nfs4_daemonproc;
+
+static int
+nfs4_do_renew(struct nfs4mount *nmp, struct ucred *cred)
+{
+	struct nfs4fs_compound compound;
+	int error;
+
+	compound.nmp = nmp;
+	compound.procp = curproc;
+	compound.cred = cred;
+
+	if ((error = nfs4_setup_compound(&compound, "nfs4_do_renew()")))
+		goto out;
+	if ((error = nfs4_setup_renew(&compound)))
+		goto nfsmout;
+
+	if ((error = nfs4fs_call_compound(&compound)))
+		goto nfsmout;
+	if ((error = nfs4_handle_simple_op(&compound, OP_RENEW, 0)))
+		goto nfsmout;
+
+ nfsmout:
+	nfs4fs_release_compound(&compound);
+ out:
+	if (error < 0)
+		error = -error;
+	return (error);
+}
+
+void
+nfs4_daemon(void *arg)
+{
+	struct mount *mp;
+	struct nfs4mount *nmp;
+
+	while (1) {
+		simple_lock(&mountlist_slock);
+		CIRCLEQ_FOREACH(mp, &mountlist, mnt_list) {
+			if (strcmp(mp->mnt_vfc->vfc_name, MOUNT_NFS) != 0)
+				continue;
+			nmp = VFSTONFS(mp);
+			if (time.tv_sec < nmp->nm_last_renewal + nmp->nm_lease_time - 4)
+				continue;
+			simple_unlock(&mountlist_slock);
+			nfs4_do_renew(nmp, (struct ucred *) arg);
+			simple_lock(&mountlist_slock);
+		}
+		simple_unlock(&mountlist_slock);
+		tsleep(&nfs4_daemonproc, PVFS, "nfs4", 2 * hz);
+	}
+}
+
 /*
  * Common code for mount and mountroot
  */
@@ -481,6 +522,7 @@ int
 mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, char *hst, struct proc *p)
 {
 	struct nfs4mount *nmp;
+	char *rpth, *cp1, *cp2;
 	int error;
 	nfsfh_t *fhp;
 	int fhsize;
@@ -488,10 +530,12 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 	struct nfs4fs_getattr getattr;
 	struct nfs4fs_getfh getfh;
 	struct nfs4fs_lookup lookup;
-	int didlookup = 0;
+	int nlookup = 0;
 	struct dentry dentry;
-	struct statfs statfsb;
+	struct nfs4fs_fattr *fp;
+	u32 max;
 
+	nfs4_setup_bitmaps();
 	if (mp->mnt_flag & MNT_UPDATE) {
 		nmp = VFSTONFS(mp);
 		/* update paths, file handles, etc, here	XXX */
@@ -503,19 +547,15 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 		bzero((caddr_t)nmp, sizeof (struct nfs4mount));
 		mp->mnt_data = (qaddr_t)nmp;
 		TAILQ_INIT(&nmp->nm_uidlruhead);
-#ifdef RPCCLNT
-		MALLOC(nmp->nm_rpcclnt, struct rpcclnt *, 
-			sizeof(struct rpcclnt), M_TEMP, M_WAITOK);
-		bzero((caddr_t)nmp->nm_rpcclnt, sizeof(struct rpcclnt));
-#endif
 	}
 
 	vfs_getnewfsid(mp);
 	nmp->nm_mountp = mp;
-	nmp->nm_timeo = NFS_TIMEO;
-	nmp->nm_retry = NFS_RETRANS;
-	nmp->nm_wsize = NFS_WSIZE;
-	nmp->nm_rsize = NFS_RSIZE;
+	nmp->nm_flag |= NFSMNT_NFSV3;
+	nmp->nm_rpcclnt.rc_timeo = NFS_TIMEO;
+	nmp->nm_rpcclnt.rc_retry = NFS_RETRANS;
+	nmp->nm_rpcclnt.rc_wsize = NFS_WSIZE;
+	nmp->nm_rpcclnt.rc_rsize = NFS_RSIZE;
 	nmp->nm_readdirsize = NFS_READDIRSIZE;
 	nmp->nm_numgrps = NFS_MAXGRPS;
 	nmp->nm_readahead = NFS_DEFRAHEAD;
@@ -529,12 +569,12 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 	bcopy(hst, mp->mnt_stat.f_mntfromname, MNAMELEN);
 	bcopy(pth, mp->mnt_stat.f_mntonname, MNAMELEN);
 	bcopy(argp, &mp->mnt_stat.mount_info.nfs_args, sizeof(*argp));
-	nmp->nm_nam = nam;
+	nmp->nm_rpcclnt.rc_name = nam;
 	nfs_decode_args(nmp, argp, &mp->mnt_stat.mount_info.nfs_args);
 
 	/* Set up the sockets and per-host congestion */
-	nmp->nm_sotype = argp->sotype;
-	nmp->nm_soproto = argp->proto;
+	nmp->nm_rpcclnt.rc_sotype = argp->sotype;
+	nmp->nm_rpcclnt.rc_soproto = argp->proto;
 
 	/*
 	 * This is silly, but it has to be set so that vinifod() works.
@@ -544,7 +584,7 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 	 */
 	mp->mnt_stat.f_iosize = NFS_MAXDGRAMDATA;
 
-	nfsv4_printk(mount, "mountnfs \"%s\"\n", (char *) argp->fh);
+	nfsv4_printk(mount, "mountnfs: hst \"%s\" pth \"%s\"\n", hst, pth);
 
 	error = nfs_connect(nmp, NULL);
 	if (error)
@@ -553,25 +593,32 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 	compound.nmp = nmp;
 	compound.procp = p;
 	compound.cred = p->p_ucred;
-	getattr.bm = &fsattr_bm;
+	getattr.bm = &fsinfo_bm;
 
 	if ((error = nfs4_setup_compound(&compound, "mountnfs()")))
 		goto out;
 	if ((error = nfs4_setup_putrootfh(&compound)))
 		goto nfsmout;
 
-	/* If mounting the server's root, skip the lookup */
-	if (strcmp((char *) argp->fh, "/") != 0) {
-		lookup.dentry = &dentry;
-		lookup.name = argp->fh;
-		if (lookup.name[0] == '/')
-			lookup.name++;
-		lookup.namelen = strlen(lookup.name);
+	/* Get remote path */
+	rpth = nfs4_strchr(hst, ':');
+	if (rpth)
+		rpth++;
+
+	compound.current_fh = lookup.dentry = &dentry;
+	for (cp1 = rpth; cp1 && *cp1; cp1 = cp2)  {
+		while (*cp1 == '/')
+			cp1++;
+		if (!*cp1)
+			break;
+		for (cp2 = cp1; *cp2 && *cp2 != '/'; cp2++)
+			;
+		lookup.name = cp1;
+		lookup.namelen = cp2 - cp1;
 		if ((error = nfs4_setup_lookup(&compound, &lookup)))
 			goto nfsmout;
-		didlookup = 1;
-	} else
-		compound.current_fh = &dentry;
+		nlookup++;
+	}
 	if ((error = nfs4_setup_getfh(&compound, &getfh)))
 		goto nfsmout;
 	if ((error = nfs4_setup_getattr(&compound, &getattr)))
@@ -581,7 +628,7 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 		goto nfsmout;
 	if ((error = nfs4_handle_putrootfh(&compound)))
 		goto nfsmout;
-	if (didlookup) {
+	while (nlookup--) {
 		if ((error = nfs4_handle_lookup(&compound)))
 			goto nfsmout;
 	}
@@ -597,11 +644,44 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 	bcopy((caddr_t) fhp, (caddr_t) nmp->nm_fh, fhsize);
 	nfsv4_printk(mount, "mountnfs fhsize %d\n", fhsize);
 
-	/* Copy filesystem attributes */
-	nfs4_stuff_fsattr(&getattr.fattr, nmp, &statfsb);
+	/* Copy fsinfo */
+	fp = &getattr.fattr;
+	if (fp->fa_valid & FA_FSID) {
+		nmp->nm_fsid.val[0] = fp->fa_fsid_major;
+		nmp->nm_fsid.val[1] = fp->fa_fsid_minor;
+	}
+	if (fp->fa_valid & FA_MAXREAD) {
+		max = fp->fa_maxread;
+		if (max < nmp->nm_rpcclnt.rc_rsize) {
+			nmp->nm_rpcclnt.rc_rsize = max & ~(NFS_FABLKSIZE - 1);
+			if (nmp->nm_rpcclnt.rc_rsize == 0)
+				nmp->nm_rpcclnt.rc_rsize = max;
+		}
+		if (max < nmp->nm_readdirsize) {
+			nmp->nm_readdirsize = max & ~(NFS_DIRBLKSIZ - 1);
+			if (nmp->nm_readdirsize == 0)
+				nmp->nm_readdirsize = max;
+		}
+	}
+	if (fp->fa_valid & FA_MAXWRITE) {
+		max = fp->fa_maxwrite;
+		if (max < nmp->nm_rpcclnt.rc_wsize) {
+			nmp->nm_rpcclnt.rc_wsize = max & ~(NFS_FABLKSIZE - 1);
+			if (nmp->nm_rpcclnt.rc_wsize == 0)
+				nmp->nm_rpcclnt.rc_wsize = max;
+		}
+	}
+	if (fp->fa_valid & FA_LEASE_TIME)
+		nmp->nm_lease_time = fp->fa_lease_time;
+
 	nmp->nm_flag |= NFSMNT_GOTFSINFO;
 
-	nfs4_do_setclientid(nmp, p);
+	if ((error = nfs4_do_setclientid(nmp, p)) != 0)
+		goto nfsmout;
+
+	/* Start renewd if it isn't already running */
+	if (nfs4_daemonproc == NULL)
+		kthread_create(nfs4_daemon, crdup(p->p_ucred), &nfs4_daemonproc, "nfs4rd");
 
  nfsmout:
 	nfs4fs_release_compound(&compound);
@@ -616,44 +696,6 @@ mountnfs(struct nfs_args *argp, struct mount *mp, struct mbuf *nam, char *pth, c
 	return (error);
 }
 
-static int
-nfs4_stuff_fsattr(struct nfs4fs_fattr *fp, struct nfs4mount *nmp, struct statfs *sbp)
-{
-	u32 max;
-
-	sbp->f_flags = nmp->nm_flag;
-	sbp->f_bsize = NFS_FABLKSIZE;
-
-	sbp->f_ffree = (fp->fa_valid & FA_FFREE) ? fp->fa_ffree : 50000;
-	sbp->f_files = (fp->fa_valid & FA_FTOTAL) ? fp->fa_ftotal : 100000;
-	if (fp->fa_valid & FA_MAXREAD) {
-		max = fp->fa_maxread;
-		if (max < nmp->nm_rsize) {
-			nmp->nm_rsize = max & ~(NFS_FABLKSIZE - 1);
-			if (nmp->nm_rsize == 0)
-				nmp->nm_rsize = max;
-		}
-		if (max < nmp->nm_readdirsize) {
-			nmp->nm_readdirsize = max & ~(NFS_DIRBLKSIZ - 1);
-			if (nmp->nm_readdirsize == 0)
-				nmp->nm_readdirsize = max;
-		}
-	}
-	if (fp->fa_valid & FA_MAXWRITE) {
-		max = fp->fa_maxwrite;
-		if (max < nmp->nm_wsize) {
-			nmp->nm_wsize = max & ~(NFS_FABLKSIZE - 1);
-			if (nmp->nm_wsize == 0)
-				nmp->nm_wsize = max;
-		}
-	}
-	sbp->f_bavail = (fp->fa_valid & FA_SAVAIL) ? (fp->fa_savail / NFS_FABLKSIZE) : 500000;
-	sbp->f_bfree = (fp->fa_valid & FA_SFREE) ? (fp->fa_sfree / NFS_FABLKSIZE) : 500000;
-	sbp->f_blocks = (fp->fa_valid & FA_STOTAL) ? (fp->fa_stotal / NFS_FABLKSIZE) : 1000000;
-
-	return 0;
-}
-
 /*
  * unmount system call
  */
@@ -663,11 +705,11 @@ nfs_unmount(mp, mntflags, p)
 	int mntflags;
 	struct proc *p;
 {
-	struct nfs4mount *nmp;
+	struct nfs4mount *nmp = VFSTONFS(mp);
 
-	nmp = VFSTONFS(mp);
+	vflush(mp, NULL, (mntflags & MNT_FORCE) ? FORCECLOSE : 0);
 	nfs_disconnect(nmp);
-	m_freem(nmp->nm_nam);
+	m_freem(nmp->nm_rpcclnt.rc_name);
 
 	return (0);
 }
@@ -680,20 +722,14 @@ nfs_root(mp, vpp)
 	struct mount *mp;
 	struct vnode **vpp;
 {
-	struct vnode *vp;
-	struct nfs4mount *nmp;
+	struct nfs4mount *nmp = VFSTONFS(mp);
 	struct nfsnode *np;
 	int error;
 
-	nmp = VFSTONFS(mp);
 	error = nfs_nget(mp, (nfsfh_t *)nmp->nm_fh, nmp->nm_fhsize, &np);
 	if (error)
 		return (error);
-	vp = NFSTOV(np);
-	if (vp->v_type == VNON)
-		vp->v_type = VDIR;
-	vp->v_flag = VROOT;
-	*vpp = vp;
+	*vpp = NFSTOV(np);
 	return (0);
 }
 
@@ -882,7 +918,8 @@ nfs4_do_setclientid(struct nfs4mount *nmp, struct proc *p)
 {
 	struct nfs4fs_compound compound;
 	struct nfs4fs_setclientid setclientid;
-	char name[30];
+	struct route ro;
+	char *ipsrc, uaddr[24], name[24];
 	int try = 0;
 	int status;
 
@@ -894,18 +931,27 @@ nfs4_do_setclientid(struct nfs4mount *nmp, struct proc *p)
 		goto out;
 	}
 
-#if 0
-	nfs4_get_callback(sb, &setclientid.callback_prog, netid, uaddr);
-#endif
 	compound.nmp = nmp;
 	compound.procp = p;
 	compound.cred = p->p_ucred;
-	setclientid.callback_netid = "udp";
-	setclientid.callback_universal_address = "141.211.133.62.12.48";
+
+	setclientid.callback_netid = (nmp->nm_rpcclnt.rc_sotype == SOCK_STREAM) ? "tcp" : "udp";
+
+	/* Do a route lookup to find our source address for talking to this server */
+	bzero(&ro, sizeof ro);
+	ro.ro_dst = *mtod(nmp->nm_rpcclnt.rc_name, struct sockaddr *);
+	rtalloc(&ro);
+	if (ro.ro_rt == NULL) {
+		status = EHOSTUNREACH;
+		goto out;
+	}
+	ipsrc = inet_ntoa(ifatoia(ro.ro_rt->rt_ifa)->ia_addr.sin_addr);
+	sprintf(uaddr, "%s.12.48", ipsrc);
+	setclientid.callback_universal_address = uaddr;
+	RTFREE(ro.ro_rt);
 
 try_again:
-	/* XXX : how do we really want to choose the name? */
-	sprintf(name, "obsd-%d", try);
+	sprintf(name, "%s-%d", ipsrc, try);
 	setclientid.nmp = nmp;
 	setclientid.namlen = strlen(name);
 	setclientid.name = name;
