@@ -1,5 +1,5 @@
-/*	$Id: nfs4_vnops.c,v 1.39 2002/09/12 19:47:24 rees Exp $	*/
-/*	$OpenBSD: nfs_vnops.c,v 1.50 2002/03/14 01:27:13 millert Exp $	*/
+/*	$Id: nfs4_vnops.c,v 1.53 2003/06/20 17:05:11 rees Exp $	*/
+/*	$OpenBSD: nfs_vnops.c,v 1.55 2003/06/02 23:28:20 millert Exp $	*/
 /*	$NetBSD: nfs_vnops.c,v 1.62.4.1 1996/07/08 20:26:52 jtc Exp $	*/
 
 /*
@@ -17,11 +17,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *	This product includes software developed by the University of
- *	California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
@@ -44,7 +40,7 @@
  * copyright (c) 2002
  * the regents of the university of michigan
  * all rights reserved
- * 
+ *
  * permission is granted to use, copy, create derivative works and redistribute
  * this software and such derivative works for any purpose, so long as the name
  * of the university of michigan is not used in any advertising or publicity
@@ -52,7 +48,7 @@
  * written prior authorization.  if the above copyright notice or any other
  * identification of the university of michigan is included in any copy of any
  * portion of this software, then the disclaimer below must also be included.
- * 
+ *
  * this software is provided as is, without representation from the university
  * of michigan as to its fitness for any purpose, and without warranty by the
  * university of michigan of any kind, either express or implied, including
@@ -85,8 +81,6 @@
 #include <sys/fcntl.h>
 #include <sys/lockf.h>
 
-#include <uvm/uvm_extern.h>
-
 #include <miscfs/specfs/specdev.h>
 #include <miscfs/fifofs/fifo.h>
 
@@ -113,7 +107,16 @@
 int nfs_openrpc(struct vnode *, struct vnode **, struct componentname *, int, struct vattr *);
 int nfs_createrpc(struct vnode *, struct vnode **, struct componentname *,
 		  nfs_ftype4, struct vattr *, char *);
-int nfs_closerpc(struct vnode *, struct ucred *, struct proc *);
+int nfs_closerpc(struct vnode *, struct ucred *, struct proc *, int);
+
+#define	nfs_lease_check ((int (*)(void *))nullop)
+#define nfs_ioctl ((int (*)(void *))enoioctl)
+#define nfs_select ((int (*)(void *))seltrue)
+#define nfs_revoke vop_generic_revoke
+#define nfs_lock ((int (*)(void *))vop_generic_lock)
+#define nfs_unlock ((int (*)(void *))vop_generic_unlock)
+#define nfs_islocked ((int (*)(void *))vop_generic_islocked)
+#define nfs_reallocblks ((int (*)(void *))eopnotsupp)
 
 /*
  * Global vfs data structures for nfs
@@ -157,7 +160,7 @@ struct vnodeopv_entry_desc nfsv2_vnodeop_entries[] = {
 	{ &vop_advlock_desc, nfs_advlock },	/* advlock */
 	{ &vop_reallocblks_desc, nfs_reallocblks },	/* reallocblks */
 	{ &vop_bwrite_desc, nfs_bwrite },
-	{ (struct vnodeop_desc*)NULL, (int(*)(void *))NULL }
+	{ NULL, NULL }
 };
 struct vnodeopv_desc nfsv2_vnodeop_opv_desc =
 	{ &nfsv2_vnodeop_p, nfsv2_vnodeop_entries };
@@ -167,44 +170,21 @@ struct vnodeopv_desc nfsv2_vnodeop_opv_desc =
  */
 int (**spec_nfsv2nodeop_p)(void *);
 struct vnodeopv_entry_desc spec_nfsv2nodeop_entries[] = {
-	{ &vop_default_desc, vn_default_error },
-	{ &vop_lookup_desc, spec_lookup },	/* lookup */
-	{ &vop_create_desc, spec_create },	/* create */
-	{ &vop_mknod_desc, spec_mknod },	/* mknod */
-	{ &vop_open_desc, spec_open },		/* open */
+	{ &vop_default_desc, spec_vnoperate },
 	{ &vop_close_desc, nfsspec_close },	/* close */
 	{ &vop_access_desc, nfsspec_access },	/* access */
 	{ &vop_getattr_desc, nfs_getattr },	/* getattr */
 	{ &vop_setattr_desc, nfs_setattr },	/* setattr */
 	{ &vop_read_desc, nfsspec_read },	/* read */
 	{ &vop_write_desc, nfsspec_write },	/* write */
-	{ &vop_lease_desc, spec_lease_check },	/* lease */
-	{ &vop_ioctl_desc, spec_ioctl },	/* ioctl */
-	{ &vop_select_desc, spec_select },	/* select */
-	{ &vop_revoke_desc, spec_revoke },	/* revoke */
 	{ &vop_fsync_desc, nfs_fsync },		/* fsync */
-	{ &vop_remove_desc, spec_remove },	/* remove */
-	{ &vop_link_desc, spec_link },		/* link */
-	{ &vop_rename_desc, spec_rename },	/* rename */
-	{ &vop_mkdir_desc, spec_mkdir },	/* mkdir */
-	{ &vop_rmdir_desc, spec_rmdir },	/* rmdir */
-	{ &vop_symlink_desc, spec_symlink },	/* symlink */
-	{ &vop_readdir_desc, spec_readdir },	/* readdir */
-	{ &vop_readlink_desc, spec_readlink },	/* readlink */
-	{ &vop_abortop_desc, spec_abortop },	/* abortop */
 	{ &vop_inactive_desc, nfs_inactive },	/* inactive */
 	{ &vop_reclaim_desc, nfs_reclaim },	/* reclaim */
 	{ &vop_lock_desc, nfs_lock },		/* lock */
 	{ &vop_unlock_desc, nfs_unlock },	/* unlock */
-	{ &vop_bmap_desc, spec_bmap },		/* bmap */
-	{ &vop_strategy_desc, spec_strategy },	/* strategy */
 	{ &vop_print_desc, nfs_print },		/* print */
 	{ &vop_islocked_desc, nfs_islocked },	/* islocked */
-	{ &vop_pathconf_desc, spec_pathconf },	/* pathconf */
-	{ &vop_advlock_desc, spec_advlock },	/* advlock */
-	{ &vop_reallocblks_desc, spec_reallocblks },	/* reallocblks */
-	{ &vop_bwrite_desc, vop_generic_bwrite },
-	{ (struct vnodeop_desc*)NULL, (int(*)(void *))NULL }
+	{ NULL, NULL }
 };
 struct vnodeopv_desc spec_nfsv2nodeop_opv_desc =
 	{ &spec_nfsv2nodeop_p, spec_nfsv2nodeop_entries };
@@ -212,44 +192,22 @@ struct vnodeopv_desc spec_nfsv2nodeop_opv_desc =
 #ifdef FIFO
 int (**fifo_nfsv2nodeop_p)(void *);
 struct vnodeopv_entry_desc fifo_nfsv2nodeop_entries[] = {
-	{ &vop_default_desc, vn_default_error },
-	{ &vop_lookup_desc, fifo_lookup },	/* lookup */
-	{ &vop_create_desc, fifo_create },	/* create */
-	{ &vop_mknod_desc, fifo_mknod },	/* mknod */
-	{ &vop_open_desc, fifo_open },		/* open */
+	{ &vop_default_desc, fifo_vnoperate },
 	{ &vop_close_desc, nfsfifo_close },	/* close */
 	{ &vop_access_desc, nfsspec_access },	/* access */
 	{ &vop_getattr_desc, nfs_getattr },	/* getattr */
 	{ &vop_setattr_desc, nfs_setattr },	/* setattr */
 	{ &vop_read_desc, nfsfifo_read },	/* read */
 	{ &vop_write_desc, nfsfifo_write },	/* write */
-	{ &vop_lease_desc, fifo_lease_check },	/* lease */
-	{ &vop_ioctl_desc, fifo_ioctl },	/* ioctl */
-	{ &vop_select_desc, fifo_select },	/* select */
-	{ &vop_revoke_desc, fifo_revoke },	/* revoke */
 	{ &vop_fsync_desc, nfs_fsync },		/* fsync */
-	{ &vop_remove_desc, fifo_remove },	/* remove */
-	{ &vop_link_desc, fifo_link },		/* link */
-	{ &vop_rename_desc, fifo_rename },	/* rename */
-	{ &vop_mkdir_desc, fifo_mkdir },	/* mkdir */
-	{ &vop_rmdir_desc, fifo_rmdir },	/* rmdir */
-	{ &vop_symlink_desc, fifo_symlink },	/* symlink */
-	{ &vop_readdir_desc, fifo_readdir },	/* readdir */
-	{ &vop_readlink_desc, fifo_readlink },	/* readlink */
-	{ &vop_abortop_desc, fifo_abortop },	/* abortop */
 	{ &vop_inactive_desc, nfs_inactive },	/* inactive */
 	{ &vop_reclaim_desc, nfs_reclaim },	/* reclaim */
 	{ &vop_lock_desc, nfs_lock },		/* lock */
 	{ &vop_unlock_desc, nfs_unlock },	/* unlock */
-	{ &vop_bmap_desc, fifo_bmap },		/* bmap */
-	{ &vop_strategy_desc, fifo_badop },	/* strategy */
 	{ &vop_print_desc, nfs_print },		/* print */
 	{ &vop_islocked_desc, nfs_islocked },	/* islocked */
-	{ &vop_pathconf_desc, fifo_pathconf },	/* pathconf */
-	{ &vop_advlock_desc, fifo_advlock },	/* advlock */
-	{ &vop_reallocblks_desc, fifo_reallocblks },	/* reallocblks */
 	{ &vop_bwrite_desc, vop_generic_bwrite },
-	{ (struct vnodeop_desc*)NULL, (int(*)(void *))NULL }
+	{ NULL, NULL }
 };
 struct vnodeopv_desc fifo_nfsv2nodeop_opv_desc =
 	{ &fifo_nfsv2nodeop_p, fifo_nfsv2nodeop_entries };
@@ -258,15 +216,9 @@ struct vnodeopv_desc fifo_nfsv2nodeop_opv_desc =
 /*
  * Global variables
  */
-extern struct nfsstats nfsstats;
 struct proc *nfs_iodwant[NFS_MAXASYNCDAEMON];
 int nfs_numasync = 0;
 
-static uint32_t __getattr_bm[] = { 0x20d00112, 0x0024be3a };
-bitmap4 getattr_bm = { 2, __getattr_bm };
-static uint32_t  __no_owner_bm[]      = { 0x00100112, 0x0024a20a };
-bitmap4          no_owner_bm          = { 2, __no_owner_bm };
-
 /*
  * nfs null call from vfs.
  */
@@ -288,9 +240,6 @@ nfs_null(vp, cred, procp)
 
 /*
  * nfs access vnode op.
- * For nfs version 2, just return ok. File accesses may fail later.
- * For nfs version 3, use the access rpc to check accessibility. If file modes
- * are changed on the server, accesses might still fail later.
  */
 int
 nfs_access(v)
@@ -406,18 +355,16 @@ nfs_openrpc(dvp, vpp, cnp, flags, vap)
 	} else {
 		/* Opening an existing file */
 		np = VTONFS(vp);
-		fp = &np->n_file_data;
+		fp = (flags & FWRITE) ? &np->n_wfile_data : &np->n_rfile_data;
 	}
 
 	/*
-	 * We only open the file once.
-	 * This is a bug and must be fixed.
+	 * We only open the file once each for reading and writing.
 	 */
 	if (fp->fi_count++) {
 		*vpp = vp;
 		goto out;
 	}
-	flags |= FWRITE;
 
 	/*
 	 * Set up file data.
@@ -437,7 +384,7 @@ nfs_openrpc(dvp, vpp, cnp, flags, vap)
 	open.vap = vap;
 	open.dentry = &dentry;
 	open.fp = fp;
-	putfh.dentry = &(VTONFS(dvp)->n_dentry);
+	putfh.dentry = &VTONFS(dvp)->n_dentry;
 	dentry.d_name.name = cnp->cn_nameptr;
 	dentry.d_name.len = cnp->cn_namelen;
 
@@ -464,13 +411,15 @@ nfs_openrpc(dvp, vpp, cnp, flags, vap)
 		vp = NFSTOV(np);
 		np->n_dvp = dvp;
 		np->n_namelen = cnp->cn_namelen;
+		if (np->n_name)
+			NFS4_FREE(np->n_name);
 		np->n_name = NFS4_ALLOC(np->n_namelen + 1);
 		memcpy(np->n_name, cnp->cn_nameptr, cnp->cn_namelen);
 		np->n_name[np->n_namelen] = '\0';
-		if (np->n_file_data.fi_count)
-			printf("nfs_openrpc(): already open %d\n", np->n_file_data.fi_count);
+		if (flags & FWRITE)
+			np->n_wfile_data = file_data;
 		else
-			np->n_file_data = file_data;
+			np->n_rfile_data = file_data;
 		error = nfs4_loadattrcache(vp, &open.getattr.fattr, NULL);
 		if (error)
 			goto nfsmout;
@@ -514,6 +463,7 @@ nfs_open(v)
 	struct nfsnode *np = VTONFS(vp);
 	struct componentname cn;
 	enum vtype vtype = vp->v_type;
+	int mode = ap->a_mode;
 
 	if (vtype != VREG) {
 		if (vtype != VDIR && vtype != VLNK) {
@@ -530,9 +480,9 @@ nfs_open(v)
 	}
 
 #ifdef DIAGNOSTIC
-	if (ap->a_mode & O_CREAT) {
+	if (mode & O_CREAT) {
 		printf("nfs_open(): attempted create\n");
-		ap->a_mode &= ~O_CREAT;
+		mode &= ~O_CREAT;
 	}
 #endif
 
@@ -541,7 +491,25 @@ nfs_open(v)
 	cn.cn_cred = ap->a_cred;
 	cn.cn_proc = ap->a_p;
 
-	return nfs_openrpc(np->n_dvp, &vp, &cn, ap->a_mode, NULL);
+	/*
+	 * Initialize read and write creds here, for swapfiles
+	 * and other paths that don't set the creds themselves.
+	 */
+
+	if (mode & FREAD) {
+		if (np->n_rcred)
+			crfree(np->n_rcred);
+		np->n_rcred = ap->a_cred;
+		crhold(np->n_rcred);
+	}
+	if (mode & FWRITE) {
+		if (np->n_wcred)
+			crfree(np->n_wcred);
+		np->n_wcred = ap->a_cred;
+		crhold(np->n_wcred);
+	}
+
+	return nfs_openrpc(np->n_dvp, &vp, &cn, mode, NULL);
 }
 
 /*
@@ -594,10 +562,7 @@ nfs_create(v)
  * nfs close rpc
  */
 int
-nfs_closerpc(vp, cred, proc)
-	struct vnode *vp;
-	struct ucred *cred;
-	struct proc *proc;
+nfs_closerpc(struct vnode *vp, struct ucred *cred, struct proc *proc, int flags)
 {
 	struct nfs4fs_file_data *fp;
 	struct nfs4fs_compound compound;
@@ -606,14 +571,11 @@ nfs_closerpc(vp, cred, proc)
 	struct nfsnode *np = VTONFS(vp);
 	int error = 0;
 
-	fp = &np->n_file_data;
+	fp = (flags & FWRITE) ? &np->n_wfile_data : &np->n_rfile_data;
 
 	if (--fp->fi_count)
 		goto out;
 
-	if (fp->fi_nopen != 1)
-		printf("nfs_close(): fp->fi_nopen == %d\n", fp->fi_nopen);
-
 	compound.nmp = VFSTONFS(vp->v_mount);
 	compound.procp = proc;
 	compound.cred = cred;
@@ -669,7 +631,7 @@ nfs_close(v)
 		np->n_attrstamp = 0;
 	}
 
-	error = nfs_closerpc(vp, ap->a_cred, ap->a_p);
+	error = nfs_closerpc(vp, ap->a_cred, ap->a_p, ap->a_fflag);
 
 	if (!error && np->n_flag & NWRITEERR) {
 		np->n_flag &= ~NWRITEERR;
@@ -814,7 +776,7 @@ nfs_setattr(v)
 		   (error = nfs_vinvalbuf(vp, V_SAVE, cred, p, 1)) == EINTR)
 		return (error);
 
-	if (vap->va_size != VNOVAL && np->n_file_data.fi_nopen == 0) {
+	if (vap->va_size != VNOVAL && np->n_wfile_data.fi_count == 0) {
 		/* Have to open the file before we can truncate it */
 		struct componentname cn;
 
@@ -850,13 +812,20 @@ nfs_setattrrpc(vp, vap, cred, procp)
 {
 	struct nfsnode *np = VTONFS(vp);
 	struct nfs4fs_compound compound;
-	struct nfs4fs_file_data *fp = &np->n_file_data;
+	struct nfs4fs_file_data *fp = &np->n_wfile_data;
 	struct nfs4fs_putfh putfh;
 	struct nfs4fs_getattr getattr;
 	int error;
 
 	nfsstats.rpccnt[NFSPROC_SETATTR]++;
 
+#ifdef DIAGNOSTIC
+	if (vap->va_size != VNOVAL && fp->fi_count < 1) {
+		printf("nfs_setattrrpc(): not open\n");
+		return EIO;
+	}
+#endif
+
 	compound.nmp = VFSTONFS(vp->v_mount);
 	compound.procp = procp;
 	compound.cred = cred;
@@ -885,7 +854,7 @@ nfs_setattrrpc(vp, vap, cred, procp)
 
 	/* TODO: do the settatr and close in a single compound rpc */
 	if (np->n_flag & NTRUNCATE) {
-		error = nfs_closerpc(vp, cred, procp);
+		error = nfs_closerpc(vp, cred, procp, FWRITE);
 		np->n_flag &= ~NTRUNCATE;
 	}
 
@@ -916,7 +885,7 @@ nfs_lookup(v)
 	struct vnode *dvp = ap->a_dvp;
 	struct vnode **vpp = ap->a_vpp;
 	struct proc *p = cnp->cn_proc;
-	int flags = cnp->cn_flags;
+	int flags;
 	struct vnode *newvp;
 	long len;
 	nfsfh_t *fhp;
@@ -929,7 +898,10 @@ nfs_lookup(v)
 	struct dentry dentry;
 	struct nfs4fs_lookup lookup;
 	struct nfs4fs_getfh getfh;
-	int didlookup;
+	int isdot;
+
+	cnp->cn_flags &= ~PDIRUNLOCK;
+	flags = cnp->cn_flags;
 
 	*vpp = NULLVP;
 	if ((flags & ISLASTCN) && (dvp->v_mount->mnt_flag & MNT_RDONLY) &&
@@ -939,51 +911,98 @@ nfs_lookup(v)
 		return (ENOTDIR);
 	lockparent = flags & LOCKPARENT;
 	wantparent = flags & (LOCKPARENT|WANTPARENT);
+	len = cnp->cn_namelen;
+	if (len == 1 && cnp->cn_nameptr[0] == '.')
+		isdot = 1;
+	else if (cnp->cn_flags & ISDOTDOT)
+		isdot = 2;
+	else
+		isdot = 0;
+
 	nmp = VFSTONFS(dvp->v_mount);
 	dnp = VTONFS(dvp);
-	if ((error = cache_lookup(dvp, vpp, cnp)) != 0 && error != ENOENT) {
+
+#ifdef NFS4DIRCACHE
+	/*
+	 * Before tediously performing a linear scan of the directory,
+	 * check the name cache to see if the directory/name pair
+	 * we are looking for is known already.
+	 * If the directory/name pair is found in the name cache,
+	 * we have to ensure the directory has not changed from
+	 * the time the cache entry has been created. If it has,
+	 * the cache entry has to be ignored.
+	 */
+	if ((error = cache_lookup(dvp, vpp, cnp)) >= 0) {
 		struct vattr vattr;
-		int vpid;
+		int err2;
+
+		if (error && error != ENOENT) {
+			*vpp = NULLVP;
+			return (error);
+		}
+
+		if (cnp->cn_flags & PDIRUNLOCK) {
+			err2 = vn_lock(dvp, LK_EXCLUSIVE | LK_RETRY, p);
+			if (err2 != 0) {
+				*vpp = NULLVP;
+				return (err2);
+			}
+			cnp->cn_flags &= ~PDIRUNLOCK;
+		}
+
+		err2 = VOP_ACCESS(dvp, VEXEC, cnp->cn_cred, cnp->cn_proc);
+		if (err2 != 0) {
+			if (error == 0) {
+				if (*vpp != dvp)
+					vput(*vpp);
+				else
+					vrele(*vpp);
+			}
+			*vpp = NULLVP;
+			return (err2);
+		}
+
+		if (error == ENOENT) {
+			if (!VOP_GETATTR(dvp, &vattr, cnp->cn_cred,
+			    cnp->cn_proc) && vattr.va_mtime.tv_sec ==
+			    VTONFS(dvp)->n_ctime)
+				return (ENOENT);
+			cache_purge(dvp);
+			np->n_ctime = 0;
+			goto dorpc;
+		}
 
 		newvp = *vpp;
-		vpid = newvp->v_id;
-		/*
-		 * See the comment starting `Step through' in ufs/ufs_lookup.c
-		 * for an explanation of the locking protocol
-		 */
-		if (dvp == newvp) {
-			VREF(newvp);
-			error = 0;
-		} else
-			error = vget(newvp, LK_EXCLUSIVE, p);
-
-		if (!error) {
-			if (vpid == newvp->v_id) {
-			   if (!VOP_GETATTR(newvp, &vattr, cnp->cn_cred, cnp->cn_proc)
-			    && vattr.va_ctime.tv_sec == VTONFS(newvp)->n_ctime) {
-				nfsstats.lookupcache_hits++;
-				if (cnp->cn_nameiop != LOOKUP &&
-				    (flags & ISLASTCN))
-					cnp->cn_flags |= SAVENAME;
-				return (0);
-			   }
-			   cache_purge(newvp);
-			}
-			vrele(newvp);
+		if (!VOP_GETATTR(newvp, &vattr, cnp->cn_cred, cnp->cn_proc)
+			&& vattr.va_ctime.tv_sec == VTONFS(newvp)->n_ctime)
+		{
+			nfsstats.lookupcache_hits++;
+			if (cnp->cn_nameiop != LOOKUP && (flags & ISLASTCN))
+				cnp->cn_flags |= SAVENAME;
+			if ((!lockparent || !(flags & ISLASTCN)) &&
+			     newvp != dvp)
+				VOP_UNLOCK(dvp, 0, p);
+			return (0);
 		}
+		cache_purge(newvp);
+		if (newvp != dvp)
+			vput(newvp);
+		else
+			vrele(newvp);
 		*vpp = NULLVP;
 	}
+dorpc:
+#endif
 	error = 0;
 	newvp = NULLVP;
 	nfsstats.lookupcache_misses++;
 	nfsstats.rpccnt[NFSPROC_LOOKUP]++;
-	len = cnp->cn_namelen;
 
 	compound.nmp = nmp;
 	compound.procp = cnp->cn_proc;
 	compound.cred = cnp->cn_cred;
-	dgetattr.bm = &no_owner_bm;
-	getattr.bm = &no_owner_bm;
+	dgetattr.bm = &getattr_bm;
+	getattr.bm = &getattr_bm;
 	putfh.dentry = &dnp->n_dentry;
 
 	if ((error = nfs4_setup_compound(&compound, "nfs_lookup()")))
@@ -992,23 +1011,20 @@ nfs_lookup(v)
 		goto nfsmout;
 	if ((error = nfs4_setup_getattr(&compound, &dgetattr)))
 		goto nfsmout;
-	if (len == 1 && cnp->cn_nameptr[0] == '.') {
+	if (isdot == 1) {
 		/* Fake up a "." lookup; nothing to do */
 		nfsv4_printk(lookup, "looking up dot\n");
-		didlookup = 0;
 	} else {
-		if (len == 2 && !strncmp(cnp->cn_nameptr, "..", 2)) {
+		if (isdot == 2) {
 			nfsv4_printk(lookup, "looking up dotdot\n");
 			if ((error = nfs4_setup_lookupp(&compound, &dentry)))
 				goto nfsmout;
-			didlookup = 1;
 		} else {
 			lookup.dentry = &dentry;
 			lookup.name = cnp->cn_nameptr;
 			lookup.namelen = len;
 			if ((error = nfs4_setup_lookup(&compound, &lookup)))
 				goto nfsmout;
-			didlookup = 2;
 		}
 		if ((error = nfs4_setup_getattr(&compound, &getattr)))
 			goto nfsmout;
@@ -1017,50 +1033,111 @@ nfs_lookup(v)
 		goto nfsmout;
 
 	if ((error = nfs4fs_call_compound(&compound)))
-		goto nfsmout;
+		goto rpcout;
 	if ((error = nfs4_handle_putfh(&compound, &putfh)))
-		goto nfsmout;
+		goto rpcout;
 	if ((error = nfs4_handle_getattr(&compound, &dgetattr)))
-		goto nfsmout;
+		goto rpcout;
 	if ((error = nfs4_loadattrcache(dvp, &dgetattr.fattr, NULL)))
 		goto nfsmout;
-	if (didlookup) {
-		if (didlookup == 1) {
+	if (isdot != 1) {
+		if (isdot == 2) {
 			if ((error = nfs4_handle_lookupp(&compound)))
-				goto nfsmout;
-		} else if (didlookup == 2) {
+				goto rpcout;
+		} else {
 			if ((error = nfs4_handle_lookup(&compound)))
-				goto nfsmout;
+				goto rpcout;
 		}
 		if ((error = nfs4_handle_getattr(&compound, &getattr)))
-			goto nfsmout;
+			goto rpcout;
 	}
 	if ((error = nfs4_handle_getfh(&compound, &getfh)))
-		goto nfsmout;
+		goto rpcout;
 	fhp = &getfh.dentry->fh_val;
 	fhsize = getfh.dentry->fh_len;
-	nfsv4_printk(lookup, "nfs_lookup %s: fhsize %d\n", cnp->cn_nameptr, fhsize);
+	nfsv4_printk(lookup, "nfs_lookup %*s: fhsize %d\n",
+		     cnp->cn_namelen, cnp->cn_nameptr, fhsize);
+
+	/*
+	 * Handle RENAME case...
+	 */
+	if (cnp->cn_nameiop == RENAME && wantparent && (flags & ISLASTCN)) {
+		if (NFS_CMPFH(dnp, fhp, fhsize)) {
+			error = EISDIR;
+			goto out;
+		}
+		error = nfs_nget(dvp->v_mount, fhp, fhsize, &np);
+		if (error) {
+			goto out;
+		}
+		newvp = NFSTOV(np);
+		np->n_dvp = dvp;
+		/*
+		 * I don't think we want to save the name in np->n_name here
+		 */
+		nfs4_loadattrcache(newvp, &getattr.fattr, NULL);
+		*vpp = newvp;
+		cnp->cn_flags |= SAVENAME;
+		if (!lockparent) {
+			VOP_UNLOCK(dvp, 0, p);
+			cnp->cn_flags |= PDIRUNLOCK;
+		}
+		goto out;
+	}
+
+	/*
+	 * The postop attr handling is duplicated for each if case,
+	 * because it should be done while dvp is locked (unlocking
+	 * dvp is different for each case).
+	 */
 
 	if (NFS_CMPFH(dnp, fhp, fhsize)) {
-		if (cnp->cn_nameiop == RENAME && wantparent && (flags & ISLASTCN)) {
-			error = EISDIR;
-			goto nfsmout;
-		}
 		VREF(dvp);
 		newvp = dvp;
+	} else if (flags & ISDOTDOT) {
+		VOP_UNLOCK(dvp, 0, p);
+		cnp->cn_flags |= PDIRUNLOCK;
+
+		error = nfs_nget(dvp->v_mount, fhp, fhsize, &np);
+		if (error) {
+			if (vn_lock(dvp, LK_EXCLUSIVE | LK_RETRY, p) == 0)
+				cnp->cn_flags &= ~PDIRUNLOCK;
+			goto out;
+		}
+		newvp = NFSTOV(np);
+
+		nfs4_loadattrcache(newvp, &getattr.fattr, NULL);
+
+		if (lockparent && (flags & ISLASTCN)) {
+			if ((error = vn_lock(dvp, LK_EXCLUSIVE, p))) {
+				vput(newvp);
+				goto out;
+			}
+			cnp->cn_flags &= ~PDIRUNLOCK;
+		}
+
 	} else {
 		error = nfs_nget(dvp->v_mount, fhp, fhsize, &np);
-		if (error)
-			goto nfsmout;
+		if (error) {
+			goto out;
+		}
 		newvp = NFSTOV(np);
 
 		/* Fill in parent vnode pointer and name (used by nfs_open) */
+		if (isdot)
+			panic("nfs_lookup: isdot");
 		np->n_dvp = dvp;
 		np->n_namelen = cnp->cn_namelen;
+		if (np->n_name)
+			NFS4_FREE(np->n_name);
 		np->n_name = NFS4_ALLOC(np->n_namelen + 1);
 		memcpy(np->n_name, cnp->cn_nameptr, cnp->cn_namelen);
 		np->n_name[np->n_namelen] = '\0';
 		nfs4_loadattrcache(newvp, &getattr.fattr, NULL);
+		if (!lockparent || !(flags & ISLASTCN)) {
+			VOP_UNLOCK(dvp, 0, p);
+			cnp->cn_flags |= PDIRUNLOCK;
+		}
 	}
 	if (cnp->cn_nameiop != LOOKUP && (flags & ISLASTCN))
 		cnp->cn_flags |= SAVENAME;
@@ -1071,14 +1148,26 @@ nfs_lookup(v)
 	}
 	*vpp = newvp;
 
-    nfsmout:
-	nfs4fs_release_compound(&compound);
-    out:
-	if (error < 0)
-		error = -error;
+ rpcout:
 	if (error) {
-		if (newvp != NULLVP)
+		/*
+		 * We get here only because of errors returned by
+		 * the RPC. Otherwise we'll have jumped to "nfsmout."
+		 */
+		if (error < 0)
+			error = -error;
+		if (error == ENOENT && (cnp->cn_flags & MAKEENTRY) &&
+		    cnp->cn_nameiop != CREATE) {
+			if (VTONFS(dvp)->n_ctime == 0)
+				VTONFS(dvp)->n_ctime =
+				    VTONFS(dvp)->n_vattr.va_mtime.tv_sec;
+			cache_enter(dvp, NULL, cnp);
+		}
+		if (newvp != NULLVP) {
 			vrele(newvp);
+			if (newvp != dvp)
+				VOP_UNLOCK(newvp, 0, p);
+		}
 		if ((cnp->cn_nameiop == CREATE || cnp->cn_nameiop == RENAME) &&
 		    (flags & ISLASTCN) && error == ENOENT) {
 			if (dvp->v_mount->mnt_flag & MNT_RDONLY)
@@ -1088,7 +1177,12 @@ nfs_lookup(v)
 		}
 		if (cnp->cn_nameiop != LOOKUP && (flags & ISLASTCN))
 			cnp->cn_flags |= SAVENAME;
+		*vpp = NULL;
 	}
+
+ nfsmout:
+	nfs4fs_release_compound(&compound);
+ out:
 	return (error);
 }
 
@@ -1181,10 +1275,9 @@ nfs_readlinkrpc(vp, uiop, cred)
  * Ditto above
  */
 int
-nfs_readrpc(vp, uiop, cred)
+nfs_readrpc(vp, uiop)
 	struct vnode *vp;
 	struct uio *uiop;
-	struct ucred *cred;
 {
 	struct nfs4mount *nmp = VFSTONFS(vp->v_mount);
 	int error = 0, len, tsiz;
@@ -1197,14 +1290,15 @@ nfs_readrpc(vp, uiop, cred)
 
 	compound.nmp = nmp;
 	compound.procp = uiop->uio_procp;
-	compound.cred = cred;
+	compound.cred = np->n_rcred;
 	compound.mreq = compound.mrep = NULL;
 	putfh.dentry = &np->n_dentry;
-	read.file_data = &np->n_file_data;
+	read.file_data = (np->n_rfile_data.fi_count) ? &np->n_rfile_data : &np->n_wfile_data;
 
 	while (tsiz > 0) {
 		nfsstats.rpccnt[NFSPROC_READ]++;
-		len = (tsiz > nmp->nm_rsize) ? nmp->nm_rsize : tsiz;
+		len = (tsiz > nmp->nm_rpcclnt.rc_rsize) ?
+			nmp->nm_rpcclnt.rc_rsize : tsiz;
 		read.offset = uiop->uio_offset;
 		read.maxcount = len;
 		read.uiop = uiop;
@@ -1241,15 +1335,14 @@ nfs_readrpc(vp, uiop, cred)
  * nfs write call
  */
 int
-nfs_writerpc(vp, uiop, cred, iomode, must_commit)
+nfs_writerpc(vp, uiop, iomode, must_commit)
 	struct vnode *vp;
 	struct uio *uiop;
-	struct ucred *cred;
 	int *iomode, *must_commit;
 {
 	struct nfs4mount *nmp = VFSTONFS(vp->v_mount);
-	int error = 0, len, tsiz, wccflag = NFSV3_WCCRATTR, rlen, commit;
-	int committed = NFSV3WRITE_FILESYNC;
+	int error = 0, len, tsiz, rlen;
+	stable_how4 commit, committed = FILE_SYNC4;
 	int32_t backup;
 	struct nfsnode *np = VTONFS(vp);
 	struct nfs4fs_compound compound;
@@ -1266,16 +1359,17 @@ nfs_writerpc(vp, uiop, cred, iomode, must_commit)
 
 	compound.nmp = nmp;
 	compound.procp = uiop->uio_procp;
-	compound.cred = cred;
+	compound.cred = np->n_wcred;
 	compound.mreq = compound.mrep = NULL;
 	putfh.dentry = &np->n_dentry;
-	write.file_data = &np->n_file_data;
+	write.file_data = &np->n_wfile_data;
 	write.stable = (stable_how4) *iomode;
 	write.uiop = uiop;
 
 	while (tsiz > 0) {
 		nfsstats.rpccnt[NFSPROC_WRITE]++;
-		len = (tsiz > nmp->nm_wsize) ? nmp->nm_wsize : tsiz;
+		len = (tsiz > nmp->nm_rpcclnt.rc_wsize) ?
+			nmp->nm_rpcclnt.rc_wsize : tsiz;
 		write.offset = uiop->uio_offset;
 		write.count = len;
 
@@ -1293,7 +1387,6 @@ nfs_writerpc(vp, uiop, cred, iomode, must_commit)
 		if ((error = nfs4_handle_write(&compound, &write)))
 			break;
 
-		wccflag = 0;
 		rlen = write.bytes_written;
 		if (rlen == 0) {
 			error = NFSERR_IO;
@@ -1312,23 +1405,19 @@ nfs_writerpc(vp, uiop, cred, iomode, must_commit)
 		 * Return the lowest committment level
 		 * obtained by any of the RPCs.
 		 */
-		if (committed == NFSV3WRITE_FILESYNC)
+		if (committed == FILE_SYNC4)
 			committed = commit;
-		else if (committed == NFSV3WRITE_DATASYNC &&
-			 commit == NFSV3WRITE_UNSTABLE)
+		else if (committed == DATA_SYNC4 &&
+			 commit == UNSTABLE4)
 			committed = commit;
 		verf = (caddr_t) write.writeverf;
 		if ((nmp->nm_flag & NFSMNT_HASWRITEVERF) == 0) {
-			bcopy(verf, (caddr_t)nmp->nm_verf, NFSX_V3WRITEVERF);
+			bcopy(verf, nmp->nm_verf, NFS4_VERIFIER_SIZE);
 			nmp->nm_flag |= NFSMNT_HASWRITEVERF;
-		} else if (bcmp(verf, (caddr_t)nmp->nm_verf, NFSX_V3WRITEVERF)) {
+		} else if (bcmp(verf, nmp->nm_verf, NFS4_VERIFIER_SIZE)) {
 			*must_commit = 1;
-			bcopy(verf, (caddr_t)nmp->nm_verf, NFSX_V3WRITEVERF);
+			bcopy(verf, nmp->nm_verf, NFS4_VERIFIER_SIZE);
 		}
-#if 0
-		if (wccflag)
-			VTONFS(vp)->n_mtime = VTONFS(vp)->n_vattr.va_mtime.tv_sec;
-#endif
 		nfs4fs_release_compound(&compound);
 		tsiz -= len;
 	}
@@ -2006,8 +2095,8 @@ nfs_readdir(v)
 		*ap->a_cookies = cookies;
 	}
 
-	if ((nmp->nm_flag & (NFSMNT_NFSV3 | NFSMNT_GOTFSINFO)) == NFSMNT_NFSV3)
-		(void)nfs_fsinfo(nmp, vp, cred, p);
+	if (!(nmp->nm_flag & NFSMNT_GOTFSINFO))
+		(void) nfs_fsinfo(nmp, vp, cred, p);
 
 	cnt = 5;
 
@@ -2127,8 +2216,6 @@ nfs_readdirrpc(struct vnode *vp,
 	struct nfsnode *dnp = VTONFS(vp);
 	int error = 0, status, tlen, more_dirs = 1, blksiz = 0, bigenough = 1;
 	unsigned int fty;
-	static u32 readdir_bmval[] = { 0x00100002 };
-	static bitmap4 readdir_bm = { 1, readdir_bmval };
 	struct nfs4fs_compound compound;
 	struct nfs4fs_putfh putfh;
 	struct nfs4fs_readdir readdir;
@@ -2236,8 +2323,10 @@ nfs_readdirrpc(struct vnode *vp,
 				     3 * NFSX_UNSIGNED);
 			cookie = fxdr_hyper(tl);
 			nfsv4_printk(readdir, "-- got cookie %d\n", (int) cookie);
+#ifdef DIAGNOSTIC
 			if (cookie < 3)
 				printf("nfs_readdirrpc: reserved cookie %d\n", (int) cookie);
+#endif
 			len = fxdr_unsigned(int, *(tl + 2));
 			if (len <= 0 || len > NFS_MAXNAMLEN) {
 				error = EBADRPC;
@@ -2343,6 +2432,29 @@ nfs_readdirrpc(struct vnode *vp,
 	return (error);
 }
 
+#ifdef __FreeBSD__
+/*
+ * Invalidate cached directory information, except for the actual directory
+ * blocks (which are invalidated separately).
+ * Done mainly to avoid the use of stale offset cookies.
+ */
+void
+nfs_invaldir(struct vnode *vp)
+{
+	struct nfsnode *np = VTONFS(vp);
+
+#ifdef DIAGNOSTIC
+	if (vp->v_type != VDIR)
+		panic("nfs: invaldir not dir");
+#endif
+	np->n_direofoffset = 0;
+	np->n_cookieverf.nfsuquad[0] = 0;
+	np->n_cookieverf.nfsuquad[1] = 0;
+	if (LIST_FIRST(&np->n_cookies))
+		LIST_FIRST(&np->n_cookies)->ndm_eocookie = 0;
+}
+#endif
+
 /*
  * Silly rename. To make the NFS filesystem that is stateless look a little
  * more like the "ufs" a remove of an active vnode is translated to a rename
@@ -2377,7 +2489,8 @@ nfs_sillyrename(dvp, vp, cnp)
 	}
 
 	/* Fudge together a funny name */
-	sp->s_namlen = sprintf(sp->s_name, ".nfsA%05x4.4", cnp->cn_proc->p_pid);
+	sp->s_namlen = snprintf(sp->s_name, sizeof sp->s_name,
+	    ".nfsA%05x4.4", cnp->cn_proc->p_pid);
 
 	/* Try lookitups until we get one that isn't there */
 	while (nfs_lookitup(dvp, sp->s_name, sp->s_namlen, sp->s_cred,
@@ -2398,7 +2511,7 @@ nfs_sillyrename(dvp, vp, cnp)
 bad:
 	vrele(sp->s_dvp);
 	crfree(sp->s_cred);
-	FREE((caddr_t)sp, M_NFSREQ);
+	FREE(sp, M_NFSREQ);
 	return (error);
 }
 
@@ -2439,7 +2552,7 @@ nfs_lookitup(dvp, name, len, cred, procp, npp)
 	lookup.name = name;
 	lookup.namelen = len;
 	lookup.dentry = &dentry;
-	getattr.bm = &no_owner_bm;
+	getattr.bm = &getattr_bm;
 
 	if ((error = nfs4_setup_compound(&compound, "nfs_lookitup()")))
 		goto out;
@@ -2469,11 +2582,11 @@ nfs_lookitup(dvp, name, len, cred, procp, npp)
 		if (*npp) {
 			np = *npp;
 			if (np->n_fhsize > NFS_SMALLFH && fhlen <= NFS_SMALLFH) {
-				free((caddr_t)np->n_fhp, M_NFSBIGFH);
+				free(np->n_fhp, M_NFSBIGFH);
 				np->n_fhp = &np->n_fh;
 			} else if (np->n_fhsize <= NFS_SMALLFH && fhlen>NFS_SMALLFH)
 				np->n_fhp =(nfsfh_t *)malloc(fhlen,M_NFSBIGFH,M_WAITOK);
-			bcopy((caddr_t)nfhp, (caddr_t)np->n_fhp, fhlen);
+			bcopy(nfhp, np->n_fhp, fhlen);
 			np->n_fhsize = fhlen;
 			newvp = NFSTOV(np);
 		} else if (NFS_CMPFH(dnp, nfhp, fhlen)) {
@@ -2488,6 +2601,8 @@ nfs_lookitup(dvp, name, len, cred, procp, npp)
 		if (newvp != dvp) {
 			np->n_dvp = dvp;
 			np->n_namelen = len;
+			if (np->n_name)
+				NFS4_FREE(np->n_name);
 			np->n_name = NFS4_ALLOC(len + 1);
 			memcpy(np->n_name, name, len);
 			np->n_name[len] = '\0';
@@ -2511,13 +2626,14 @@ nfs_lookitup(dvp, name, len, cred, procp, npp)
 }
 
 /*
- * Nfs Version 3 commit rpc
+ * nfs commit rpc
  */
 int
-nfs_commit(vp, offset, cnt, procp)
+nfs_commit(vp, offset, cnt, cred, procp)
 	struct vnode *vp;
 	u_quad_t offset;
 	int cnt;
+	struct ucred *cred;
 	struct proc *procp;
 {
 	struct nfs4mount *nmp = VFSTONFS(vp->v_mount);
@@ -2525,19 +2641,19 @@ nfs_commit(vp, offset, cnt, procp)
 	struct nfs4fs_putfh putfh;
 	struct nfs4fs_commit commit;
 	int error = 0;
-	
+
 	if ((nmp->nm_flag & NFSMNT_HASWRITEVERF) == 0)
 		return (0);
 	nfsstats.rpccnt[NFSPROC_COMMIT]++;
 
 	compound.nmp = nmp;
 	compound.procp = procp;
-	compound.cred = procp->p_ucred;
-	putfh.dentry = &(VTONFS(vp)->n_dentry);
+	compound.cred = cred;
+	putfh.dentry = &VTONFS(vp)->n_dentry;
 	commit.start = offset;
 	commit.length = cnt;
 
-	if ((error = nfs4_setup_compound(&compound, "nfs_commitrpc()")))
+	if ((error = nfs4_setup_compound(&compound, "nfs_commit()")))
 		goto out;
 	if ((error = nfs4_setup_putfh(&compound, &putfh)))
 		goto nfsmout;
@@ -2551,8 +2667,8 @@ nfs_commit(vp, offset, cnt, procp)
 	if ((error = nfs4_handle_commit(&compound, &commit)))
 		goto nfsmout;
 
-	if (bcmp((caddr_t) nmp->nm_verf, commit.verifier, NFS4_VERIFIER_SIZE)) {
-		bcopy(commit.verifier, (caddr_t) nmp->nm_verf, NFS4_VERIFIER_SIZE);
+	if (bcmp(nmp->nm_verf, commit.verifier, NFS4_VERIFIER_SIZE)) {
+		bcopy(commit.verifier, nmp->nm_verf, NFS4_VERIFIER_SIZE);
 		error = NFSERR_STALEWRITEVERF;
 	}
 
@@ -2682,7 +2798,7 @@ nfs_flush(vp, cred, waitfor, p, commit)
 	 */
 again:
 	bvecpos = 0;
-	if (NFS_ISV3(vp) && commit) {
+	if (commit) {
 		s = splbio();
 		for (bp = LIST_FIRST(&vp->v_dirtyblkhd); bp != NULL; bp = nbp) {
 			nbp = LIST_NEXT(bp, b_vnbufs);
@@ -2715,7 +2831,7 @@ again:
 		/*
 		 * Commit data on the server, as required.
 		 */
-		retv = nfs_commit(vp, off, (int)(endoff - off), p);
+		retv = nfs_commit(vp, off, (int)(endoff - off), cred, p);
 		if (retv == NFSERR_STALEWRITEVERF)
 			nfs_clearcommit(vp->v_mount);
 		/*
@@ -2735,8 +2851,8 @@ again:
 			    bp->b_flags |= B_ASYNC;
 			    bp->b_flags &= ~(B_READ|B_DONE|B_ERROR);
 			    bp->b_dirtyoff = bp->b_dirtyend = 0;
-			    splx(s);
 			    biodone(bp);
+			    splx(s);
 			}
 		}
 	}
@@ -2814,9 +2930,7 @@ loop:
 
 /*
  * Return POSIX pathconf information applicable to nfs.
- *
- * The NFS V2 protocol doesn't support this, so just return EINVAL
- * for V2.
+ * We could do this, but it's not high on my list.
  */
 /* ARGSUSED */
 int
@@ -2867,8 +2981,8 @@ nfs_print(v)
 	struct vnode *vp = ap->a_vp;
 	struct nfsnode *np = VTONFS(vp);
 
-	printf("tag VT_NFS, fileid %ld fsid 0x%lx",
-		np->n_vattr.va_fileid, np->n_vattr.va_fsid);
+	printf("tag VT_NFS, \"%*s\" fileid %ld fsid 0x%lx",
+		np->n_namelen, np->n_name, np->n_vattr.va_fileid, np->n_vattr.va_fsid);
 #ifdef FIFO
 	if (vp->v_type == VFIFO)
 		fifo_printinfo(vp);
@@ -2932,13 +3046,15 @@ nfs_writebp(bp, force)
 	if ((oldflags & (B_NEEDCOMMIT | B_WRITEINPROG)) == B_NEEDCOMMIT) {
 		off = ((u_quad_t)bp->b_blkno) * DEV_BSIZE + bp->b_dirtyoff;
 		bp->b_flags |= B_WRITEINPROG;
-		retv = nfs_commit(bp->b_vp, off, bp->b_dirtyend-bp->b_dirtyoff,
-			bp->b_proc);
+		retv = nfs_commit(bp->b_vp, off, bp->b_dirtyend - bp->b_dirtyoff,
+			VTONFS(bp->b_vp)->n_wcred, bp->b_proc);
 		bp->b_flags &= ~B_WRITEINPROG;
 		if (!retv) {
 			bp->b_dirtyoff = bp->b_dirtyend = 0;
 			bp->b_flags &= ~B_NEEDCOMMIT;
+			s = splbio();
 			biodone(bp);
+			splx(s);
 		} else if (retv == NFSERR_STALEWRITEVERF)
 			nfs_clearcommit(bp->b_vp->v_mount);
 	}
@@ -2955,7 +3071,7 @@ nfs_writebp(bp, force)
 		}
 		brelse(bp);
 		return (rtval);
-	} 
+	}
 
 	return (0);
 }
