@@ -1,4 +1,4 @@
-/*	$Id: nfs4fs.h,v 1.26 2002/08/20 21:49:22 rees Exp $	*/
+/*	$Id: nfs4fs.h,v 1.31 2003/03/02 02:28:06 rees Exp $	*/
 
 /*
  * copyright (c) 2002
@@ -187,28 +187,11 @@ struct nfs4fs_lockowner {
 struct nfs4fs_file_data {
 	struct nfsnode			*fi_n;
 	pid_t				fi_pid;
-/*	struct files_struct		*fi_fs;*/
 
 	atomic_t			fi_count;	/* refcount */
 	struct nfs4fs_lockowner	*	fi_lockowner;	/* constant */
 	struct list_head		fi_list;	/* all files on this lockowner */
 
-	/*
-	 * For open stateids only:
-	 *  These are lists of all OPEN's for read (resp. write).  These are linked
-	 *  either through the inode or the delegation, depending on whether the
-	 *  OPEN is a "real" open or a delegated one.
-	 */
-	struct list_head		fi_read;
-	struct list_head		fi_write;
-
-	/*
-	 * These are leftover from a previous state model
-	 */
-	unsigned int			fi_nopen;
-	unsigned int			fi_nread;
-	unsigned int			fi_nwrite;
-
 	/*
 	 * Linkage for open_stateids and lock_stateids.
 	 * There is another subtlety here: it would be nice if the lock_stateid could just
@@ -496,23 +479,11 @@ struct nfs4fs_read {
 };
 
 struct nfs4fs_readdir {
-	/* XXX : for now, READDIRs must be done _last_ within a COMPOUND to work! */
 	/* filled in by caller */
 	u32 count;
 	bitmap4 *bm;
-	filldir_t filler;
-	void *buf;   /* opaque argument to filldir_t */
-	
-	/* filled in by caller, but changed by handler */
-	/* XXX : right now, the cookie is htonl'ed on send, ntohl'ed on receive.  we
-	   should change things so that the cookie is unchanged, since it is opaque to
-	   the client anyway.. */
 	nfs_cookie4 cookie;
 	verifier4 verifier;
-
-	/* filled in by handler */
-	int eof;
-	int eob;
 };
 
 struct nfs4fs_rename {
@@ -530,9 +501,11 @@ struct nfs4fs_setclientid {
 	char *callback_netid;
 	char *callback_universal_address;
 
-	/* filled in by handler */
+	/* filled in by handler, used by setclientid_confirm */
 	int in_use;
+	int old_proto;
 	clientid4 clientid;
+	verifier4 verifier;
 };
 
 struct nfs4fs_write {
@@ -677,6 +650,7 @@ extern int nfs4_setup_readdir(struct nfs4fs_compound *cp, struct nfs4fs_readdir
 #define    nfs4_setup_readlink(cp)     nfs4_setup_simple_op((cp), OP_READLINK)
 extern int nfs4_setup_remove(struct nfs4fs_compound *cp, char *name, int namelen);
 extern int nfs4_setup_rename(struct nfs4fs_compound *cp, struct nfs4fs_rename *rename);
+extern int nfs4_setup_renew(struct nfs4fs_compound *cp);
 extern int nfs4_setup_restorefh(struct nfs4fs_compound *cp);
 extern int nfs4_setup_reval(struct nfs4fs_compound *cp, struct nfs4fs_getattr *reval);
 extern int nfs4_setup_savefh(struct nfs4fs_compound *cp);
@@ -685,7 +659,8 @@ extern int nfs4_setup_setattr(struct nfs4fs_compound *cp, struct vattr *vap,
 extern int nfs4_setup_setattr_mode(struct nfs4fs_compound *cp, umode_t mode);
 extern int nfs4_setup_setclientid(struct nfs4fs_compound *cp,
 				  struct nfs4fs_setclientid *setclientid);
-extern int nfs4_setup_setclientid_confirm(struct nfs4fs_compound *cp, clientid4 clientid);
+extern int nfs4_setup_setclientid_confirm(struct nfs4fs_compound *cp,
+					  struct nfs4fs_setclientid *setclientid);
 extern int nfs4_setup_write(struct nfs4fs_compound *cp, struct nfs4fs_write *write);
 extern bitmap4 nfs4_positive_bitmap;
 extern bitmap4 nfs4_negative_bitmap;
@@ -765,6 +740,9 @@ extern void nfs4fs_release_compound(struct nfs4fs_compound *cp);
 extern struct rpc_program nfs4_program;
 extern struct rpc_stat nfs4_rpcstat;
 
-int nfs4_loadattrcache(struct vnode *vp, struct nfs4fs_fattr *fp, struct vattr *vaper);
+/* subs.c */
+extern bitmap4 fsinfo_bm, fsattr_bm, getattr_bm, readdir_bm;
+extern int nfs4_loadattrcache(struct vnode *vp, struct nfs4fs_fattr *fp, struct vattr *vaper);
+extern void nfs4_setup_bitmaps(void);
 
 #endif
