@@ -1,4 +1,4 @@
-/*	$Id: nfs4fs_decode.c,v 1.4 2002/09/12 17:36:59 rees Exp $	*/
+/*	$Id: nfs4fs_decode.c,v 1.7 2003/03/04 02:02:40 rees Exp $	*/
 
 /*
  * copyright (c) 2002
@@ -51,6 +51,7 @@
 #include <nfs/nfs.h>
 #include <nfs4/list.h>
 #include <nfs4/nfs4_prot.h>
+#include <nfs4/nfs4_gss.h>
 #include <nfs4/nfs4_debug.h>
 #include <nfs4/nfs4.h>
 #include <nfs4/nfs4mount.h>
@@ -205,6 +206,7 @@ nfs4_read_attrs(struct nfs4fs_compound *cp, struct nfs4fs_fattr *nfp)
 	u32 *p;
 	int status;
 	int attrlen, len, xdrlen;
+	u_int32_t id;
 
 	/* XXX : need more sanity checking on attributes read! */
 	if (!(p = read_buf(cp, 4)))
@@ -253,8 +255,8 @@ nfs4_read_attrs(struct nfs4fs_compound *cp, struct nfs4fs_fattr *nfp)
 		READ_BUF(16);
 		READ64(nfp->fa_fsid_major);
 		READ64(nfp->fa_fsid_minor);
-		nfsv4_printk(xdr, "read_attrs: fsid=0x%Lx/0x%Lx\n",
-			     nfp->fa_fsid_major, nfp->fa_fsid_minor);
+		nfsv4_printk(xdr, "read_attrs: fsid=0x%x/0x%x\n",
+			     (u32)nfp->fa_fsid_major, (u32)nfp->fa_fsid_minor);
 		nfp->fa_valid |= FA_FSID;
 	}
 	if (FATTR4_ISSET(FATTR4_LEASE_TIME, bmval)) {
@@ -274,21 +276,21 @@ nfs4_read_attrs(struct nfs4fs_compound *cp, struct nfs4fs_fattr *nfp)
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_ffree);
-		nfsv4_printk(xdr, "read_attrs: files_free=0x%Lx\n", nfp->fa_ffree);
+		nfsv4_printk(xdr, "read_attrs: files_free=0x%x\n", (int) nfp->fa_ffree);
 		nfp->fa_valid |= FA_FFREE;
 	}
 	if (FATTR4_ISSET(FATTR4_FILES_TOTAL, bmval)) {
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_ftotal);
-		nfsv4_printk(xdr, "read_attrs: files_tot=0x%Lx\n", nfp->fa_ftotal);
+		nfsv4_printk(xdr, "read_attrs: files_tot=0x%x\n", (int) nfp->fa_ftotal);
 		nfp->fa_valid |= FA_FTOTAL;
 	}
 	if (FATTR4_ISSET(FATTR4_MAXFILESIZE, bmval)) {
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_maxfilesize);
-		nfsv4_printk(getattr, "read_attrs: maxfilesize=0x%Lx\n", nfp->fa_maxfilesize);
+		nfsv4_printk(getattr, "read_attrs: maxfilesize=0x%x\n", (int) nfp->fa_maxfilesize);
 		nfp->fa_valid |= FA_MAXFILESIZE;
 	}
 	if (FATTR4_ISSET(FATTR4_MAXNAME, bmval)) {
@@ -302,14 +304,14 @@ nfs4_read_attrs(struct nfs4fs_compound *cp, struct nfs4fs_fattr *nfp)
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_maxread);
-		nfsv4_printk(xdr, "read_attrs: maxread=0x%Lx\n", nfp->fa_maxread);
+		nfsv4_printk(xdr, "read_attrs: maxread=0x%x\n", (int) nfp->fa_maxread);
 		nfp->fa_valid |= FA_MAXREAD;
 	}
 	if (FATTR4_ISSET(FATTR4_MAXWRITE, bmval)) {
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_maxwrite);
-		nfsv4_printk(xdr, "read_attrs: maxwrite=0x%Lx\n", nfp->fa_maxwrite);
+		nfsv4_printk(xdr, "read_attrs: maxwrite=0x%x\n", (int) nfp->fa_maxwrite);
 		nfp->fa_valid |= FA_MAXWRITE;
 	}
 
@@ -336,30 +338,30 @@ nfs4_read_attrs(struct nfs4fs_compound *cp, struct nfs4fs_fattr *nfp)
 		OVERFLOW_CHECK(4);
 		READ_BUF(4);
 		len = ntohl(*p++);
-		nfsv4_printk(xdr, "read_attrs: owner len=%d\n", len);
 		if (len > GSS_NAME_MAX)
 			goto xdr_error;
 		xdrlen = XDR_LEN(len);
 		OVERFLOW_CHECK(xdrlen);
 		READ_BUF(xdrlen);
-		if ((status = gss_get_num(GSS_OWNER, len, (char *)p, &nfp->fa_uid)))
+		if ((status = gss_string_to_uint32(GSS_OWNER, len, (char *)p, &id)))
 			goto out;
-		nfsv4_printk(xdr, "read_attrs: uid=%d\n", (int)nfp->fa_uid);
+		nfsv4_printk(xdr, "read_attrs: uid=%*s(%d)\n", len, (char *)p, id);
+		nfp->fa_uid = id;
 		nfp->fa_valid |= FA_UID;
 	}
 	if (FATTR4_ISSET(FATTR4_OWNER_GROUP, bmval)) {
 		OVERFLOW_CHECK(4);
 		READ_BUF(4);
 		len = ntohl(*p++);
-		nfsv4_printk(xdr, "read_attrs: group len=%d\n", len);
 		if (len > GSS_NAME_MAX)
 			goto xdr_error;
 		xdrlen = XDR_LEN(len);
 		OVERFLOW_CHECK(xdrlen);
 		READ_BUF(xdrlen);
-		if ((status = gss_get_num(GSS_GROUP, len, (char *)p, &nfp->fa_gid)))
+		if ((status = gss_string_to_uint32(GSS_GROUP, len, (char *)p, &id)))
 			goto out;
-		nfsv4_printk(xdr, "read_attrs: gid=%d\n", (int)nfp->fa_gid);
+		nfsv4_printk(xdr, "read_attrs: gid=%*s(%d)\n", len, (char *)p, id);
+		nfp->fa_gid = id;
 		nfp->fa_valid |= FA_GID;
 	}
 	if (FATTR4_ISSET(FATTR4_RAWDEV, bmval)) {
@@ -375,21 +377,21 @@ nfs4_read_attrs(struct nfs4fs_compound *cp, struct nfs4fs_fattr *nfp)
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_savail);
-		nfsv4_printk(xdr, "read_attrs: savail=0x%Lx\n", nfp->fa_savail);
+		nfsv4_printk(xdr, "read_attrs: savail=0x%x\n", (int) nfp->fa_savail);
 		nfp->fa_valid |= FA_SAVAIL;
 	}
 	if (FATTR4_ISSET(FATTR4_SPACE_FREE, bmval)) {
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_sfree);
-		nfsv4_printk(xdr, "read_attrs: sfree=0x%Lx\n", nfp->fa_sfree);
+		nfsv4_printk(xdr, "read_attrs: sfree=0x%x\n", (int) nfp->fa_sfree);
 		nfp->fa_valid |= FA_SFREE;
 	}
 	if (FATTR4_ISSET(FATTR4_SPACE_TOTAL, bmval)) {
 		OVERFLOW_CHECK(8);
 		READ_BUF(8);
 		READ64(nfp->fa_stotal);
-		nfsv4_printk(xdr, "read_attrs: stotal=0x%Lx\n", nfp->fa_stotal);
+		nfsv4_printk(xdr, "read_attrs: stotal=0x%x\n", (int) nfp->fa_stotal);
 		nfp->fa_valid |= FA_STOTAL;
 	}
 	if (FATTR4_ISSET(FATTR4_SPACE_USED, bmval)) {
@@ -593,7 +595,6 @@ nfs4_handle_close(struct nfs4fs_compound *rp, struct nfs4fs_close *close)
 
 	if (copy_buf(rp, sizeof(stateid4), file_data->fi_stateid))
 		goto xdr_error;
-	file_data->fi_nopen--;
 
 out:
 	return status;
@@ -668,14 +669,6 @@ xdr_error:
 	goto out;
 }
 
-/* attributes which are required on a GETATTR.
-   XXX : right now, we require a lot of attributes to be returned.. later, think about how
-   to relax this requirement. */
-#define POSITIVE_MASK	(FA_TYPE | FA_FSID | FA_FILEID | FA_SIZE | FA_MODE | FA_NLINK | \
-			 FA_UID | FA_GID | FA_RDEV | FA_ATIME | FA_MTIME)
-#define NEGATIVE_MASK	(FA_TYPE | FA_FSID | FA_FILEID | FA_SIZE | FA_MODE | FA_NLINK | \
-			 FA_UID | FA_GID | FA_RDEV | FA_ATIME | FA_CTIME | FA_MTIME)
-
 int
 nfs4_handle_getattr(struct nfs4fs_compound *cp, struct nfs4fs_getattr *getattr)
 {
@@ -895,11 +888,6 @@ nfs4_handle_open(struct nfs4fs_compound *cp, struct nfs4fs_open *open)
 		nfsv4_printk(open, "nfs4_handle_open: confirm flag is not set\n");
 		nfs4_up_write(&fp->fi_sema);
 		up(&open->lockowner->lo_sema);
-		fp->fi_nopen++;
-		if (open->flags & FREAD)
-			fp->fi_nread++;
-		if (open->flags & FWRITE)
-			fp->fi_nwrite++;
 	}
 	else {
 		/*
@@ -1011,12 +999,6 @@ nfs4_handle_open_confirm(struct nfs4fs_compound *rp, struct nfs4fs_open *open)
 	if (copy_buf(rp, sizeof(stateid4), fp->fi_stateid))
 		goto xdr_error;
 
-	fp->fi_nopen++;
-	if (open->flags & FREAD)
-		fp->fi_nread++;
-	if (open->flags & FWRITE)
-		fp->fi_nwrite++;
-
 	status = 0;
 
 out:
