@@ -1,4 +1,4 @@
-/*	$Id: nfs4fs_encode.c,v 1.4 2002/09/12 17:36:59 rees Exp $	*/
+/*	$Id: nfs4fs_encode.c,v 1.9 2003/03/04 02:02:40 rees Exp $	*/
 
 /*
  * copyright (c) 2002
@@ -51,6 +51,7 @@
 #include <nfs/nfs.h>
 #include <nfs4/list.h>
 #include <nfs4/nfs4_prot.h>
+#include <nfs4/nfs4_gss.h>
 #include <nfs4/nfs4_debug.h>
 #include <nfs4/nfs4.h>
 #include <nfs4/nfs4mount.h>
@@ -140,15 +141,6 @@ adjust_args(struct nfs4fs_compound *cp, u32 *p)
  *  Setup routines
  */
 
-/* type, changeid, size, fsid, fileid */
-static u32 positive_bmval[] = { 0x0010011a };
-bitmap4 nfs4_positive_bitmap = { 1, positive_bmval };
-
-/* type, changeid, size, fsid, fileid, mode, nlink, owner, group,
-   rawdev, ctime, atime, mtime */
-static u32 negative_bmval[] = { 0x0010011a, 0x0020823a };
-bitmap4 nfs4_negative_bitmap = { 2, negative_bmval };
-
 int
 nfs4_setup_compound(struct nfs4fs_compound *cp, char *tag)
 {
@@ -206,8 +198,8 @@ encode_attrs(struct nfs4fs_compound *cp, struct vattr *vap, fattr4_acl *acl)
 {
 	u32 *p, *q, *bmval, *attrlenp;
 	int len, bmlen = 2;
-	struct gss_cacheent *owner = NULL;
-	struct gss_cacheent *group = NULL;
+	struct gss_name *owner = NULL;
+	struct gss_name *group = NULL;
 	int status;
 #ifdef NFS4ACL
 	nfsace4 *ace;
@@ -223,18 +215,18 @@ encode_attrs(struct nfs4fs_compound *cp, struct vattr *vap, fattr4_acl *acl)
 	 */
 	len = 36;
 	if (vap->va_uid != VNOVAL) {
-		if ((status = gss_get_name(GSS_OWNER, vap->va_uid, &owner))) {
+		if ((status = gss_uint32_to_string(GSS_OWNER, vap->va_uid, &owner))) {
 			printk("nfs4: couldn't resolve uid %d to string\n", vap->va_uid);
 			goto out;
 		}
-		len += (XDR_LEN(owner->name.len) + 4);
+		len += (XDR_LEN(owner->len) + 4);
 	}
 	if (vap->va_gid != VNOVAL) {
-		if ((status = gss_get_name(GSS_GROUP, vap->va_gid, &group))) {
+		if ((status = gss_uint32_to_string(GSS_GROUP, vap->va_gid, &group))) {
 			printk("nfs4: couldn't resolve uid %d to string\n", vap->va_gid);
 			goto out;
 		}
-		len += (XDR_LEN(group->name.len) + 4);
+		len += (XDR_LEN(group->len) + 4);
 	}
 #ifdef NFS4ACL
 	if (acl != NULL) {
@@ -285,13 +277,13 @@ encode_attrs(struct nfs4fs_compound *cp, struct vattr *vap, fattr4_acl *acl)
 	}
 	if (vap->va_uid != VNOVAL) {
 		FATTR4_SET(FATTR4_OWNER, bmval);
-		WRITE32(owner->name.len);
-		WRITEMEM(owner->name.name, owner->name.len);
+		WRITE32(owner->len);
+		WRITEMEM(owner->name, owner->len);
 	}
 	if (vap->va_gid != VNOVAL) {
 		FATTR4_SET(FATTR4_OWNER_GROUP, bmval);
-		WRITE32(group->name.len);
-		WRITEMEM(group->name.name, group->name.len);
+		WRITE32(group->len);
+		WRITEMEM(group->name, group->len);
 	}
 	if (vap->va_atime.tv_sec != VNOVAL) {
 		FATTR4_SET(FATTR4_TIME_ACCESS_SET, bmval);
@@ -316,9 +308,9 @@ encode_attrs(struct nfs4fs_compound *cp, struct vattr *vap, fattr4_acl *acl)
 	status = 0;
 out:
 	if (owner)
-		gss_put(owner);
+		gss_free_name(owner);
 	if (group)
-		gss_put(group);
+		gss_free_name(group);
 	return status;
 nomem:
 	status = -ENOMEM;
@@ -373,44 +365,18 @@ nfs4_setup_close(struct nfs4fs_compound *cp, struct nfs4fs_close *close)
 {
 	u32 *p;
 	struct nfs4fs_file_data *fdata = close->data;
-	int access;
 	int status;
 
-#if NFS4_DEBUG
-	if (!fdata || !fdata->fi_nopen) {
-		printk("!!! nfs4_setup_close: no opened file !!!\n");
-		return -EIO;
-	}
-#endif
-
 	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
 		goto out;
 	if (!(p = reserve_space(cp, 10)))
 		goto nomem;
 
-	if (fdata->fi_nopen == 1) {
-		nfsv4_printk(level2, "    setting up op #%d: CLOSE\n", cp->req_nops);
-		*p++ = close->op = htonl(OP_CLOSE);
-		cp->seqidp = p;
-		p++;
-		cp->stateidp[cp->nstateids++] = p;
-		p += XDR_QUADLEN(sizeof(stateid4));
-	}
-	else {
-		access = 0;
-		if (fdata->fi_nread > close->read)
-			access |= OPEN4_SHARE_ACCESS_READ;
-		if (fdata->fi_nwrite > close->write)
-			access |= OPEN4_SHARE_ACCESS_WRITE;
-		nfsv4_printk(level2, "    setting up op #%d: OPEN_DOWNGRADE\n", cp->req_nops);
-		*p++ = close->op = htonl(OP_OPEN_DOWNGRADE);
-		cp->stateidp[cp->nstateids++] = p;
-		p += XDR_QUADLEN(sizeof(stateid4));
-		cp->seqidp = p;
-		p++;
-		*p++ = htonl(access);
-		*p++ = htonl(OPEN4_SHARE_DENY_NONE);
-	}
+	nfsv4_printk(level2, "    setting up op #%d: CLOSE\n", cp->req_nops);
+	*p++ = close->op = htonl(OP_CLOSE);
+	cp->seqidp = p++;
+	cp->stateidp[cp->nstateids++] = p;
+	p += XDR_QUADLEN(sizeof(stateid4));
 	adjust_args(cp, p);
 	cp->req_nops++;
 	cp->flags = (CA_STATEFUL | CA_SEQID_RELEASE);
@@ -503,7 +469,7 @@ nfs4_setup_create(struct nfs4fs_compound *cp, struct nfs4fs_create *create)
 	cp->req_nops++;
 
 	create->getattr.flags = FL_NONFATAL;
-	create->getattr.bm = &nfs4_negative_bitmap;
+	create->getattr.bm = &getattr_bm;
 	create->getfh.nonfatal = 1;
 	if ((status = nfs4_setup_getattr(cp, &create->getattr)))
 		goto out;
@@ -752,7 +718,7 @@ nfs4_setup_open(struct nfs4fs_compound *cp, struct nfs4fs_open *open)
 	cp->renew_index = cp->req_nops;
 
 	open->getattr.flags = FL_CLOBBER_OK;
-	open->getattr.bm = &nfs4_negative_bitmap;
+	open->getattr.bm = &getattr_bm;
 	open->getfh.nonfatal = 1;
 	if ((status = nfs4_setup_getattr(cp, &open->getattr)))
 		goto out;
@@ -944,6 +910,30 @@ nomem:
 	goto out;
 }
 
+
+int
+nfs4_setup_renew(struct nfs4fs_compound *cp)
+{
+	u32 *p;
+	int status;
+
+	if (!(p = reserve_space(cp, 3)))
+		goto nomem;
+	WRITE32(OP_RENEW);
+	WRITE64(cp->nmp->nm_clientid);
+	adjust_args(cp, p);
+	cp->req_nops++;
+	cp->renew_index = cp->req_nops;
+	status = 0;
+
+out:
+	return status;
+nomem:
+	nfsv4_printk(xdr2, "setup_renew: out of memory!\n");
+	status = -ENOMEM;
+	goto out;
+}
+
 int
 nfs4_setup_restorefh(struct nfs4fs_compound *cp)
 {
