@@ -97,7 +97,7 @@
 
 #define RPC_ERRSTR_ACCEPTED_SIZE 6
 char * rpc_errstr_accepted[RPC_ERRSTR_ACCEPTED_SIZE] = {
-"0xDEADBEEF",
+"", /* no good message... */
 "remote server hasn't exported program.",
 "remote server can't support version number.",
 "program can't support procedure.",
@@ -112,12 +112,12 @@ char * rpc_errstr_denied[2] = {
 
 #define RPC_ERRSTR_AUTH_SIZE 6
 char * rpc_errstr_auth[RPC_ERRSTR_AUTH_SIZE] = {
-"authentication ok???",
-"authentication error: bad credential (seal broken).",
-"authentication error: client must begin new session.",
-"authentication error: bad verifier (seal broken).",
-"authentication error: verifier expired or replayed.",
-"authentication error: rejected for security reasons.",
+"",
+"auth error: bad credential (seal broken).",
+"auth error: client must begin new session.",
+"auth error: bad verifier (seal broken).",
+"auth error: verifier expired or replayed.",
+"auth error: rejected for security reasons.",
 };
 
 
@@ -140,6 +140,7 @@ struct rpcrtt rpcrtt;
  */
 TAILQ_HEAD(, rpctask) rpctask_q;
 struct pool rpctask_pool;
+struct pool rpcclnt_pool;
 #define RPCTASKPOOL_LWM 10
 #define RPCTASKPOOL_HWM 40
 
@@ -195,10 +196,12 @@ rpcclnt_init()
 	/* initialize rpctask */
 	TAILQ_INIT(&rpctask_q);
 	pool_init(&rpctask_pool, sizeof(struct rpctask), 0, 0, RPCTASKPOOL_LWM,
-			"rpctask pool", NULL);
+			"rpctask_p", NULL);
 	pool_setlowat(&rpctask_pool, RPCTASKPOOL_LWM);
 	pool_sethiwat(&rpctask_pool, RPCTASKPOOL_HWM);
 
+	pool_init(&rpcclnt_pool, sizeof(struct rpcclnt), 0, 0, 1, "rpcclnt_p", NULL);
+
 	/* initialize timers */
 	timeout_set(&rpcclnt_timer_to, rpcclnt_timer, &rpcclnt_timer_to);
 	rpcclnt_timer(&rpcclnt_timer_to);
@@ -206,12 +209,9 @@ rpcclnt_init()
 	return;
 }
 
-/*
- * Initialize sockets and congestion for a new RPC connection.
- * We do not free the sockaddr if error.
- */
+/* wrapper arround rpcclnt_connect() that sets more options: */
 int
-rpcclnt_connect(rpc, flag, nam, sotype, soproto, progid, progvers, servername,
+rpcclnt_setup(rpc, flag, nam, sotype, soproto, progid, progvers, servername,
 		rsize, wsize)
 	struct rpcclnt *rpc;
 	int	flag;		/* rpc specific flags carry over from mount.h
@@ -225,14 +225,8 @@ rpcclnt_connect(rpc, flag, nam, sotype, soproto, progid, progvers, servername,
 	int 	rsize;
 	int 	wsize;
 {
-	struct socket *so;
-	int s, error, rcvreserve, sndreserve;
-	struct sockaddr *saddr;
-	struct sockaddr_in *sin;
-	struct mbuf *m;
-
-	if(rpc == NULL){
-		RPCDEBUG("no rpcclnt struct!\n");
+	if (rpc == NULL) {
+		RPCDEBUG("null rpc struct passed");
 		return EFAULT;
 	}
 
@@ -245,23 +239,44 @@ rpcclnt_connect(rpc, flag, nam, sotype, soproto, progid, progvers, servername,
 	rpc->rc_soproto = soproto;
 	rpc->rc_progid = progid;
 	rpc->rc_progvers = progvers;
-	if(rpc->rc_servername != (char *)NULL){
-                FREE(rpc->rc_servername, M_TEMP);
-        }
-	MALLOC(rpc->rc_servername, char *, strlen(servername)+1, M_TEMP, 
-		M_WAITOK);
-	kcopy(servername, rpc->rc_servername, strlen(servername)+1);
+	rpc->rc_servername = servername;
  	rpc->rc_rtton = 0;
         rpc->rc_proctlen = 0;
-        rpc->rc_proct = (int *)NULL;
+        rpc->rc_proct = NULL;
 
+	return rpcclnt_connect(rpc);
+}
 
-	rpc->rc_so = (struct socket *)NULL;
+/*
+ * Initialize sockets and congestion for a new RPC connection.
+ * We do not free the sockaddr if error.
+ */
+int
+rpcclnt_connect(rpc)
+	struct rpcclnt *rpc;
+{
+	struct socket *so;
+	int s, error, rcvreserve, sndreserve;
+	struct sockaddr *saddr;
+	struct sockaddr_in *sin;
+	struct mbuf *m;
+
+	if(rpc == NULL){
+		RPCDEBUG("no rpcclnt struct!\n");
+		return EFAULT;
+	}
+
+	/* create the socket */
+	rpc->rc_so = NULL;
 	saddr = mtod(rpc->rc_name, struct sockaddr *);
+
 	error = socreate(saddr->sa_family, &rpc->rc_so, rpc->rc_sotype, 
 		rpc->rc_soproto);
-	if (error)
-		goto bad;
+
+	if (error) {
+		RPCDEBUG("error %d in socreate()", error);
+		return error;
+	}
 	so = rpc->rc_so;
 	rpc->rc_soflags = so->so_proto->pr_flags;
 
@@ -390,19 +405,6 @@ bad:
 	return (error);
 }
 
-void
-rpcclnt_destroy(rpc)
-	struct rpcclnt * rpc;
-{
-	if(rpc->rc_servername != NULL)
-		FREE(rpc->rc_servername, M_TEMP);
-	if(rpc->rc_authstr != NULL)
-		FREE(rpc->rc_authstr, M_TEMP);
-	if(rpc->rc_verfstr != NULL)
-		FREE(rpc->rc_verfstr, M_TEMP);
-	
-	return;
-}
 
 /*
  * Reconnect routine:
@@ -418,17 +420,14 @@ rpcclnt_reconnect(rep)
 	struct rpctask *rep;
 {
 	struct rpctask *rp;
-	struct rpcclnt *rpc= rep->r_rpcclnt;
+	struct rpcclnt *rpc = rep->r_rpcclnt;
 	int error;
 
 	rpcclnt_disconnect(rpc);
-	while ((error = rpcclnt_connect(rpc, rpc->rc_flag, rpc->rc_name, 
-			rpc->rc_sotype, rpc->rc_soproto, rpc->rc_progid,
-			rpc->rc_progvers, rpc->rc_servername, rpc->rc_rsize,
-			rpc->rc_wsize)) != 0) {
+	while ((error = rpcclnt_connect(rpc)) != 0) {
 		if (error == EINTR || error == ERESTART)
 			return (EINTR);
-		(void) tsleep((caddr_t)&lbolt, PSOCK, "rpccon", 0);
+		tsleep(&lbolt, PSOCK, "rpccon", 0);
 	}
 
 	/*
@@ -454,7 +453,7 @@ rpcclnt_disconnect(rpc)
 
 	if (rpc->rc_so) {
 		so = rpc->rc_so;
-		rpc->rc_so = (struct socket *)0;
+		rpc->rc_so = NULL;
 		soshutdown(so, 2);
 		soclose(so);
 	}
@@ -498,17 +497,19 @@ rpcclnt_send(so, nam, top, rep)
 		soflags = rep->r_rpcclnt->rc_soflags;
 	} else
 		soflags = so->so_proto->pr_flags;
+
 	if ((soflags & PR_CONNREQUIRED) || (so->so_state & SS_ISCONNECTED))
-		sendnam = (struct mbuf *)0;
+		sendnam = NULL;
 	else
 		sendnam = nam;
+
 	if (so->so_type == SOCK_SEQPACKET)
 		flags = MSG_EOR;
 	else
 		flags = 0;
 
-	error = sosend(so, sendnam, (struct uio *)0, top,
-		(struct mbuf *)0, flags);
+	error = sosend(so, sendnam, NULL, top, NULL, flags);
+
 	if (error) {
 		if (rep) {
 			log(LOG_INFO, "rpc send error %d for server %s\n",error,
@@ -524,7 +525,7 @@ rpcclnt_send(so, nam, top, rep)
 			log(LOG_INFO, "rpc service send error %d\n", error);
 
 		/*
-		 * Handle any recoverable (soft) socket errors here. (???)
+		 * Handle any recoverable (soft) socket errors here. 
 		 */
 		if (error != EINTR && error != ERESTART &&
 			error != EWOULDBLOCK && error != EPIPE)
@@ -561,8 +562,8 @@ rpcclnt_receive(rep, aname, mp)
 	/*
 	 * Set up arguments for soreceive()
 	 */
-	*mp = (struct mbuf *)0;
-	*aname = (struct mbuf *)0;
+	*mp = NULL;
+	*aname = NULL;
 	sotype = rep->r_rpcclnt->rc_sotype;
 
 	/*
@@ -625,21 +626,20 @@ tryagain:
 			auio.uio_resid = sizeof(u_int32_t);
 			auio.uio_procp = p;
 			do {
-			   rcvflg = MSG_WAITALL;
-			   error = soreceive(so, (struct mbuf **)0, &auio,
-				(struct mbuf **)0, (struct mbuf **)0, &rcvflg);
-			   if (error == EWOULDBLOCK && rep) {
-				if (rep->r_flags & R_SOFTTERM)
-					return (EINTR);
-			   }
+				rcvflg = MSG_WAITALL;
+				error = soreceive(so, NULL, &auio, NULL, NULL, &rcvflg);
+				if (error == EWOULDBLOCK && rep) {
+					if (rep->r_flags & R_SOFTTERM)
+						return (EINTR);
+				}
 			} while (error == EWOULDBLOCK);
 			if (!error && auio.uio_resid > 0) {
-			    log(LOG_INFO,
-				 "short receive (%d/%d) from rpc server %s\n",
-				 sizeof(u_int32_t) - auio.uio_resid,
-				 sizeof(u_int32_t),
-				 rep->r_rpcclnt->rc_servername);
-			    error = EPIPE;
+				log(LOG_INFO,
+				    "short receive (%d/%d) from rpc server %s\n",
+				    sizeof(u_int32_t) - auio.uio_resid,
+				    sizeof(u_int32_t),
+				    rep->r_rpcclnt->rc_servername);
+				error = EPIPE;
 			}
 			if (error)
 				goto errout;
@@ -649,26 +649,26 @@ tryagain:
 			 * and forcing a disconnect/reconnect is all I can do.
 			 */
 			if (len > RPC_MAXPACKET) {
-			    log(LOG_ERR, "%s (%d) from rpc server %s\n",
-				"impossible packet length",
-				len,
-				rep->r_rpcclnt->rc_servername);
-			    error = EFBIG;
-			    goto errout;
+				log(LOG_ERR, "%s (%d) from rpc server %s\n",
+				    "impossible packet length",
+				    len,
+				    rep->r_rpcclnt->rc_servername);
+				error = EFBIG;
+				goto errout;
 			}
 			auio.uio_resid = len;
 			do {
-			    rcvflg = MSG_WAITALL;
-			    error =  soreceive(so, (struct mbuf **)0,
-				&auio, mp, (struct mbuf **)0, &rcvflg);
+				rcvflg = MSG_WAITALL;
+				error =  soreceive(so, (struct mbuf **)0,
+						   &auio, mp, (struct mbuf **)0, &rcvflg);
 			} while (error == EWOULDBLOCK || error == EINTR ||
 				 error == ERESTART);
 			if (!error && auio.uio_resid > 0) {
-			    log(LOG_INFO,
-				"short receive (%d/%d) from rpc server %s\n",
-				len - auio.uio_resid, len,
-				rep->r_rpcclnt->rc_servername);
-			    error = EPIPE;
+				log(LOG_INFO,
+				    "short receive (%d/%d) from rpc server %s\n",
+				    len - auio.uio_resid, len,
+				    rep->r_rpcclnt->rc_servername);
+				error = EPIPE;
 			}
 		} else {
 			/*
@@ -1117,14 +1117,28 @@ rpcclnt_request(rpc, mrest, procnum, procp, reply)
 
 		reply->verf_type = fxdr_unsigned(u_int32_t, *tl++);
 		reply->verf_size = fxdr_unsigned(u_int32_t, *tl);
-		rpcm_adv(rpcm_rndup(reply->verf_size));
+
+		if (reply->verf_size != 0)
+			rpcm_adv(rpcm_rndup(reply->verf_size));
 		
 		rpcm_dissect(tl, u_int32_t *, RPCX_UNSIGNED);
 		reply->stat.status = fxdr_unsigned(u_int32_t, *tl);
 
 		if (reply->stat.status == RPC_SUCCESS) {
-			reply->result_md = md;
-			reply->result_dpos = dpos;
+			if ((int)(dpos - mtod(md, caddr_t)) >= md->m_len) {
+				RPCDEBUG("where is the next mbuf?");
+				RPCDEBUG("%d -> %d", (int)(dpos-mtod(md,caddr_t)), md->m_len);
+				if (md->m_next == NULL) {
+					error = EBADRPC;	
+					goto rpcmout;
+				} else {
+					reply->result_md = md->m_next;
+					reply->result_dpos = mtod(reply->result_md, caddr_t);
+				}
+			} else {
+				reply->result_md = md;
+				reply->result_dpos = dpos;
+			}
 		} else if (reply->stat.status == RPC_PROGMISMATCH) {
 			rpcm_dissect(tl, u_int32_t *, 2 * RPCX_UNSIGNED);
 			reply->stat.mismatch_info.low = fxdr_unsigned(u_int32_t, *tl++);
@@ -1467,7 +1481,8 @@ rpcclnt_realign(m, hsiz)
 int
 rpcclnt_msg(p, server, msg)
 	struct proc *p;
-	char *server, *msg;
+	const char *server;
+	char *msg;
 {
 	tpr_t tpr;
 
@@ -1475,7 +1490,7 @@ rpcclnt_msg(p, server, msg)
 		tpr = tprintf_open(p);
 	else
 		tpr = NULL;
-	tprintf(tpr, "nfs server %s: %s\n", server, msg);
+	tprintf(tpr, "rpc server %s: %s\n", server, msg);
 	tprintf_close(tpr);
 	return (0);
 }
@@ -1501,21 +1516,34 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 	register caddr_t bpos;
 	register int i;
 	struct mbuf *mreq, *mb2;
-	int siz, grpsiz, authsiz;	
+	size_t siz, grpsiz, authsiz, verfsiz;	
+	char * tmpstr;
+
+	/* for null verf, type and size(=0) are sent */
+	verfsiz = 2 * RPCX_UNSIGNED;
 
 	/*XXX:  check authsize!!!! */
-	if(rc->rc_authtype == RPCAUTH_NULL){
+	if(rc->rc_authtype == RPCAUTH_NULL) {
 		authsiz = 0;
+	} else if (rc->rc_authtype == RPCAUTH_UNIX) {
+		if (rc->rc_cr == NULL)
+			authsiz = 5 * RPCX_UNSIGNED;
+		else 
+			authsiz = (5 + rc->rc_cr->cr_ngroups) * RPCX_UNSIGNED;
 	} else {
+		verfsiz = rpcm_rndup(rc->rc_verflen) + (2 * RPCX_UNSIGNED);
 		authsiz = rpcm_rndup(rc->rc_authlen);
 	}
 
+	/* 8 RPCX_UNSIGNED used before next call */
+
 	MGETHDR(mb, M_WAIT, MT_DATA);
-	if ((authsiz + 10 * RPCX_UNSIGNED) >= MINCLSIZE) {
+	if (authsiz + (8 * RPCX_UNSIGNED) + verfsiz >= MINCLSIZE) {
 		MCLGET(mb, M_WAIT);
-	} else if ((authsiz + 10 * RPCX_UNSIGNED) < MHLEN) {
-		MH_ALIGN(mb, authsiz + 10 * RPCX_UNSIGNED);
+	} else if (authsiz + (8 * RPCX_UNSIGNED) + verfsiz < MHLEN) {
+		MH_ALIGN(mb, authsiz + (8 * RPCX_UNSIGNED));
 	} else {
+		/* XXX */
 		MH_ALIGN(mb, 8 * RPCX_UNSIGNED);
 	}
 	mb->m_len = 0;
@@ -1554,19 +1582,29 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 	*tl = txdr_unsigned(authsiz);
 	switch (rc->rc_authtype) {
 	case RPCAUTH_UNIX:
-		rpcm_build(tl, u_int32_t *, rc->rc_authlen);
-		*tl++ = 0;              /* stamp ?? */
+		rpcm_build(tl, u_int32_t *, authsiz);
 
+		*tl++ = 0;              /* stamp ?? */
 		*tl++ = 0;              /* NULL hostname */
-		*tl++ = txdr_unsigned(rc->rc_cr->cr_uid);
-		*tl++ = txdr_unsigned(rc->rc_cr->cr_gid);
-		grpsiz = (rc->rc_authlen >> 2) - 5;
-		*tl++ = txdr_unsigned(grpsiz);
-		for (i = 0; i < grpsiz; i++)
-		*tl++ = txdr_unsigned(rc->rc_cr->cr_groups[i]);
+
+		if (rc->rc_cr != NULL) {
+			*tl++ = txdr_unsigned(rc->rc_cr->cr_uid);
+			*tl++ = txdr_unsigned(rc->rc_cr->cr_gid);
+			//grpsiz = (rc->rc_authlen >> 2) - 5;
+			grpsiz = rc->rc_cr->cr_ngroups;
+			*tl++ = txdr_unsigned(grpsiz);
+			for (i = 0; i < grpsiz; i++)
+				*tl++ = txdr_unsigned(rc->rc_cr->cr_groups[i]);
+		} else {
+			RPCDEBUG("using root creds");
+			*tl++ = 0;
+			*tl++ = 0;
+			*tl++ = 0;
+		}
 	break;
 	case RPCAUTH_KERB4:
 		siz = rc->rc_authlen;
+		tmpstr = rc->rc_authstr;
 		while (siz > 0) {
 			if (M_TRAILINGSPACE(mb) == 0) {
 				MGET(mb2, M_WAIT, MT_DATA);
@@ -1578,9 +1616,9 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 				bpos = mtod(mb, caddr_t);
 			}
 			i = min(siz, M_TRAILINGSPACE(mb));
-			bcopy(rc->rc_authstr, bpos, i);
+			bcopy(tmpstr, bpos, i);
 			mb->m_len += i;
-			rc->rc_authstr += i;
+			tmpstr += i;
 			bpos += i;
 			siz -= i;
 		}
@@ -1594,8 +1632,9 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 		/* nothing to be done */
 	break;
 	default:
-		panic("rpcclnt_buildheader - auth doesnt exist = %d\n", 
-			rc->rc_authtype);
+		RPCDEBUG("auth type doesnt exist = %d\n", rc->rc_authtype);
+		m_freem(mb);
+		return NULL;
 	break;
 	};
 
@@ -1607,6 +1646,7 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 		*tl++ = txdr_unsigned(RPCAUTH_KERB4);
 		*tl = txdr_unsigned(rc->rc_verflen);
 		siz = rc->rc_verflen;
+		tmpstr = rc->rc_verfstr;
 		while (siz > 0) {
 			if (M_TRAILINGSPACE(mb) == 0) {
 				MGET(mb2, M_WAIT, MT_DATA);
@@ -1618,9 +1658,9 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 				bpos = mtod(mb, caddr_t);
 			}
 			i = min(siz, M_TRAILINGSPACE(mb));
-			bcopy(rc->rc_verfstr, bpos, i);
+			bcopy(tmpstr, bpos, i);
 			mb->m_len += i;
-			rc->rc_verfstr += i;
+			tmpstr += i;
 			bpos += i;
 			siz -= i;
 		}
@@ -1630,14 +1670,16 @@ rpcclnt_buildheader(rc, procid, mrest, mrest_len, xidp, mheadend)
 			mb->m_len += siz;
 		}
 	} else {
+		RPCDEBUG("using nil verifier");
 		*tl++ = txdr_unsigned(RPCAUTH_NULL);
 		*tl = 0;
 	}
 
 	*mheadend = mb;
 	mb->m_next = mrest;
-	mreq->m_pkthdr.len = authsiz + 10 * RPCX_UNSIGNED + mrest_len;
-	mreq->m_pkthdr.rcvif = (struct ifnet *)0;
+	mreq->m_pkthdr.len = authsiz + (8 * RPCX_UNSIGNED) + verfsiz 
+				+ mrest_len;
+	mreq->m_pkthdr.rcvif = NULL;
 	return (mreq);
 }
 
@@ -1733,6 +1775,9 @@ rpcm_disct(mdp, dposp, siz, left, cp2)
 	return (0);
 }
 
+
+
+/* if auth_type == RPCAUTH_UNIX && cred == NULL, then root creds are used */
 void
 rpcclnt_setauth(rpc, auth_type, auth_len, auth_str, verf_len, verf_str, cred)
 	struct rpcclnt * rpc;
@@ -1750,28 +1795,6 @@ rpcclnt_setauth(rpc, auth_type, auth_len, auth_str, verf_len, verf_str, cred)
 	rpc->rc_authstr = auth_str;
 	rpc->rc_verfstr = verf_str;
 
-	#if 0
-	if(rpc->rc_authstr != NULL){
-		FREE(rpc->rc_authstr, M_TEMP);		
-	}
-	if(auth_len != 0){
-		MALLOC(rpc->rc_authstr, char *, auth_len + 1, M_TEMP,
-			M_WAITOK);
-		kcopy(auth_str, rpc->rc_authstr, auth_len);
-		rpc->rc_authstr[auth_len] = '\0';
-	}
-
-	if(rpc->rc_verfstr != NULL){
-		FREE(rpc->rc_verfstr, M_TEMP);		
-	}
-	if(verf_len != 0){
-		MALLOC(rpc->rc_verfstr, char *, verf_len + 1, M_TEMP,
-			M_WAITOK);
-		kcopy(verf_str, rpc->rc_verfstr, verf_len);
-		rpc->rc_verfstr[verf_len] = '\0';
-	}
-	#endif
-
 	rpc->rc_cr = cred;
 }
 
@@ -1844,3 +1867,4 @@ int rpcclnt_err(reply)
 		return EACCES;
 	}
 }
+
